"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING
from typing import Type

from encommon.types import DictStrAny
from encommon.types import NCNone

from .child import RobieChild
from ..addons import RobieQueue
from ...utils import ClientChannels

if TYPE_CHECKING:
    from ..models import RobieCommand
    from ..models import RobieMessage
    from ..params import RobieClientParams
    from ..threads import RobieClientThread



class RobieClient(RobieChild):
    """
    Establish and maintain connection with the chat service.
    """

    __thread: Optional['RobieClientThread'] = None


    @property
    def family(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        raise NotImplementedError


    @property
    def kind(
        self,
    ) -> Literal['client']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'client'


    @property
    def channels(
        self,
    ) -> ClientChannels:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        raise NotImplementedError


    @classmethod
    def schema(
        cls,
    ) -> Type['RobieClientParams']:
        """
        Return the configuration parameters relevant for class.

        :returns: Configuration parameters relevant for class.
        """

        raise NotImplementedError


    @property
    def thread(
        self,
    ) -> Optional['RobieClientThread']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        from ..threads import (
            RobieClientThread)

        thread = self.__thread

        if thread is None:
            return NCNone

        assert isinstance(
            thread,
            RobieClientThread)

        return thread


    @thread.setter
    def thread(
        self,
        value: 'RobieClientThread',
    ) -> None:
        """
        Update the value for the attribute from class instance.
        """

        from ..threads import (
            RobieClientThread)

        assert isinstance(
            value,
            RobieClientThread)

        self.__thread = value


    @property
    def dumped(
        self,
    ) -> DictStrAny:
        """
        Return the facts about the attributes from the instance.

        :returns: Facts about the attributes from the instance.
        """

        dumped = super().dumped

        family = self.family

        return dumped | {
            'family': family}


    def operate(
        self,
    ) -> None:
        """
        Perform the operation related to Robie service threads.

        .. note::
           Deviates from enhomie in children have operations,
           and are more isolated from internal core routines.
        """

        raise NotImplementedError


    def get_message(
        self,
    ) -> 'RobieMessage':
        """
        Return the new item containing information for operation.

        :returns: New item containing information for operation.
        """

        raise NotImplementedError


    def put_message(
        self,
        mqueue: RobieQueue['RobieMessage'],
    ) -> None:
        """
        Insert the new item containing information for operation.

        :param mqueue: Queue instance where the item is received.
        """

        raise NotImplementedError


    def get_command(
        self,
    ) -> 'RobieCommand':
        """
        Return the new item containing information for operation.

        :returns: New item containing information for operation.
        """

        raise NotImplementedError


    def put_command(
        self,
        cqueue: RobieQueue['RobieCommand'],
    ) -> None:
        """
        Insert the new item containing information for operation.

        :param cqueue: Queue instance where the item is received.
        """

        raise NotImplementedError


    def compose(
        self,
        target: str,
        content: str,
    ) -> 'RobieCommand':
        """
        Compose the message and transmit using the chat client.

        :param target: Where the composed message will be sent.
        :param content: Content that will be source of message.
        """

        raise NotImplementedError



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from .thread import RobieThread

if TYPE_CHECKING:
    from ..childs import RobiePlugin
    from ..members import RobiePlugins



class RobiePluginThread(RobieThread):
    """
    Common methods and routines for Chatting Robie threads.
    """


    @property
    def member(
        self,
    ) -> 'RobiePlugins':
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        from ..members import (
            RobiePlugins)

        member = super().member

        assert isinstance(
            member, RobiePlugins)

        return member


    @property
    def child(
        self,
    ) -> 'RobiePlugin':
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        from ..childs import (
            RobiePlugin)

        child = super().child

        assert isinstance(
            child, RobiePlugin)

        return child


    @property
    def plugin(
        self,
    ) -> 'RobiePlugin':
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.child



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from typing import Any
from typing import Optional
from typing import TYPE_CHECKING
from typing import Type

from encommon.types import DictStrAny
from encommon.types import NCNone
from encommon.types import clsname
from encommon.utils import array_ansi
from encommon.utils import print_ansi

from .addons import RobieJinja2
from .addons import RobieLogger
from .childs import RobieChilds
from .childs import RobieClient
from .childs import RobiePlugin
from .models import RobieModels
from ..utils import importer

if TYPE_CHECKING:
    from .common import RobiePrint
    from .config import RobieConfig
    from .params import RobieParams
    from .childs import RobiePerson



class Robie:
    """
    Interact with chat networks and integrate using plugins.

    :param config: Primary class instance for configuration.
    """

    __config: 'RobieConfig'

    __logger: RobieLogger
    __jinja2: RobieJinja2

    __childs: RobieChilds


    def __init__(
        self,
        config: 'RobieConfig',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        config.logger.log_d(
            base=clsname(self),
            status='initial')

        self.__config = config

        self.__logger = (
            RobieLogger(self))

        self.__jinja2 = (
            RobieJinja2(self))

        self.__childs = (
            RobieChilds(self))

        self.childs.validate()

        config.logger.log_d(
            base=clsname(self),
            status='created')

        self.register_locate()


    @property
    def config(
        self,
    ) -> 'RobieConfig':
        """
        Return the Config instance containing the configuration.

        :returns: Config instance containing the configuration.
        """

        return self.__config


    @property
    def logger(
        self,
    ) -> RobieLogger:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__logger


    @property
    def jinja2(
        self,
    ) -> RobieJinja2:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__jinja2


    @property
    def childs(
        self,
    ) -> RobieChilds:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__childs


    @property
    def params(
        self,
    ) -> 'RobieParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self.config.params


    @property
    def dumped(
        self,
    ) -> DictStrAny:
        """
        Return the facts about the attributes from the instance.

        :returns: Facts about the attributes from the instance.
        """

        params = deepcopy(
            self.params.endumped)

        childs = deepcopy(
            self.childs.dumped)

        items = childs.items()

        for key, value in items:
            params[key] = value

        return params


    def printer(
        self,
        source: 'RobiePrint',
        color: int = 6,
    ) -> None:
        """
        Print the contents for the object within Robie instance.

        :param source: Content which will be shown after header.
        :param color: Override the color used for box character.
        """

        params = self.params
        printer = params.printer

        pmessage = printer.message
        pcommand = printer.command


        RobieMessage = (
            RobieModels
            .message())

        RobieCommand = (
            RobieModels
            .command())


        message = isinstance(
            source, RobieMessage)

        if message and not pmessage:
            return NCNone


        command = isinstance(
            source, RobieCommand)

        if command and not pcommand:
            return NCNone


        line: str = '━'

        print_ansi(
            f'\n<c9{color}>┍'
            f'{line * 63}<c0>')

        dumped = array_ansi(
            source, indent=2)

        print(  # noqa: T201
            f'\n{dumped}')

        print_ansi(
            f'\n<c9{color}>┕'
            f'{line * 63}<c0>\n')


    def register(
        self,
        name: str,
        *,
        client: Optional[Type['RobieClient']] = None,
        plugin: Optional[Type['RobiePlugin']] = None,
    ) -> None:
        """
        Register the plugin with the internal operation routine.

        :param name: Name of the object within the Robie config.
        :param client: Class definition for the instantiation.
        :param plugin: Class definition for the instantiation.
        """

        self.childs.register(
            name=name,
            client=client,
            plugin=plugin)


    def register_locate(
        self,
    ) -> None:
        """
        Register the plugin with the internal operation routine.
        """

        config = self.config
        merge = config.merge


        clients = (
            merge.get('clients')
            or {})

        items = clients.items()

        for name, source in items:

            locate = (
                source
                .get('locate'))

            if locate is None:
                continue  # NOCVR

            client = (
                importer(locate))

            assert issubclass(
                client,
                RobieClient)

            self.register(
                name=name,
                client=client)


        plugins = (
            merge.get('plugins')
            or {})

        items = plugins.items()

        for name, source in items:

            locate = (
                source
                .get('locate'))

            if locate is None:
                continue  # NOCVR

            plugin = (
                importer(locate))

            assert issubclass(
                plugin,
                RobiePlugin)

            self.register(
                name=name,
                plugin=plugin)


    def person(
        self,
        client: RobieClient,
        check: str,
    ) -> Optional['RobiePerson']:
        """
        Return the heaviest weighted match using provided check.

        :param client: Client class instance for Chatting Robie.
        :param check: Value to be searched within the haystack.
        :returns: Heaviest weighted match using provided check.
        """

        childs = self.childs

        persons = (
            childs.persons
            .values())


        matched: set['RobiePerson'] = set()


        for person in persons:

            if not person.enable:
                continue  # NOCVR

            match = person.match(
                client, check)

            if match is True:
                matched.add(person)


        if len(matched) == 0:
            return None

        ordered = sorted(
            matched,
            key=lambda x: x.weight,
            reverse=True)

        return ordered[0]


    def j2parse(
        self,
        value: Any,  # noqa: ANN401
        statics: Optional[DictStrAny] = None,
        literal: bool = True,
    ) -> Any:  # noqa: ANN401
        """
        Return the provided input using the Jinja2 environment.

        :param value: Input that will be processed and returned.
        :param statics: Additional values available for parsing.
        :param literal: Determine if Python objects are evaled.
        :returns: Provided input using the Jinja2 environment.
        """

        return self.__jinja2.parse(
            value, statics, literal)



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from time import sleep as block_sleep
from typing import TYPE_CHECKING
from typing import Type

from encommon.times import Timer
from encommon.types import NCNone
from encommon.types.strings import SPACED

from .params import AutoJoinPluginParams
from ..status import StatusPlugin
from ..status import StatusPluginStates
from ...robie.childs import RobiePlugin

if TYPE_CHECKING:
    from ...clients.irc.message import IRCMessage



_CHANNELS = dict[str, set[str]]



class AutoJoinPlugin(RobiePlugin):
    """
    Integrate with the Robie routine and perform operations.

    .. note::
       This plugin maintains joined for configured channels.
    """

    __started: bool

    __should: _CHANNELS
    __joined: _CHANNELS

    __timer: Timer


    def __post__(
        self,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__started = False

        params = self.params


        should: _CHANNELS = {}
        joined: _CHANNELS = {}

        channels = params.channels

        for channel in channels:

            name = channel.client

            if name not in should:
                should[name] = set()
                joined[name] = set()

            should[name].add(
                channel.channel)

        self.__should = should
        self.__joined = joined


        self.__timer = Timer(
            params.interval)

        self.__status('pending')


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        # Review the parameters


    @classmethod
    def schema(
        cls,
    ) -> Type[AutoJoinPluginParams]:
        """
        Return the configuration parameters relevant for class.

        :returns: Configuration parameters relevant for class.
        """

        return AutoJoinPluginParams


    @property
    def params(
        self,
    ) -> AutoJoinPluginParams:
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        params = super().params

        assert isinstance(
            params,
            AutoJoinPluginParams)

        return params


    def operate(
        self,
    ) -> None:
        """
        Perform the operation related to Robie service threads.
        """

        from ...clients.irc.message import IRCMessage

        assert self.thread

        thread = self.thread
        mqueue = thread.mqueue
        timer = self.__timer

        clients = (
            thread.service
            .clients.childs)


        if not self.__started:
            self.__started = True
            self.__status('normal')


        if timer.ready():
            self.__operate()


        while not mqueue.empty:

            mitem = mqueue.get()

            name = mitem.client
            family = mitem.family

            # Ignore unrelated clients
            if family != 'irc':
                continue

            assert isinstance(
                mitem, IRCMessage)

            # Ignore disabled clients
            if name not in clients:
                continue  # NOCVR

            self.__message(mitem)


    def __operate(
        self,
    ) -> None:
        """
        Perform the operation related to Robie service threads.
        """

        from ...clients import IRCClient

        assert self.thread

        thread = self.thread
        should = self.__should
        joined = self.__joined
        member = thread.member
        cqueue = member.cqueue

        clients = (
            thread.service
            .clients.childs)

        failure: set[bool] = set()


        for name in should:

            # Ignore disabled clients
            if name not in clients:
                return NCNone

            client = clients[name]

            assert isinstance(
                client, IRCClient)

            connected = (
                client.client
                .connected)

            # Bypass when disconnected
            if connected is False:
                failure.add(True)
                return None

            _should = should[name]
            _joined = joined[name]

            for join in _should:

                if join in _joined:
                    continue  # NOCVR

                failure.add(True)

                rawcmd = f'JOIN :{join}'

                client.put_command(
                    cqueue, rawcmd)


        self.__status(
            'failure'
            if any(failure)
            else 'normal')


    def __message(
        self,
        mitem: 'IRCMessage',
    ) -> None:
        """
        Process the provided message item from the Robie thread.

        :param mitem: Item containing information for operation.
        """

        from ...clients import IRCClient

        robie = self.robie
        childs = robie.childs
        clients = childs.clients
        joined = self.__joined

        _client = mitem.client
        isme = mitem.isme
        event = mitem.event

        client = clients[_client]

        if _client not in joined:
            return NCNone

        assert isinstance(
            client, IRCClient)

        _joined = joined[_client]


        command = event.command
        params = event.params

        if params is None:
            return NCNone

        split = params.split(SPACED)

        current = (
            client.client
            .nickname)


        if command == '376':

            _joined.clear()

            # Allow for identify
            block_sleep(5)

            self.__operate()


        elif (command == 'JOIN'
                and isme is True):

            channel = split[0][1:]

            _joined.add(channel)


        elif command == 'KICK':

            channel = split[0]
            target = split[1]

            if target == current:
                _joined.remove(channel)


        elif (command == 'PART'
                and isme is True):

            channel = split[0]

            _joined.remove(channel)


    def __status(
        self,
        status: StatusPluginStates,
    ) -> None:
        """
        Update or insert the status of the Robie child instance.

        :param status: One of several possible value for status.
        """

        thread = self.thread
        params = self.params

        if thread is None:
            return None

        plugins = (
            thread.service
            .plugins.childs)

        if 'status' not in plugins:
            return NCNone

        plugin = plugins['status']

        assert isinstance(
            plugin, StatusPlugin)

        (plugin.update(
            unique=self.name,
            group='Management',
            title='IRC',
            icon=params.status,
            state=status))



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from dataclasses import dataclass
from typing import Literal
from typing import TYPE_CHECKING

from enconnect.mattermost import ClientEvent

from ...robie.models import RobieMessage

if TYPE_CHECKING:
    from ...robie import Robie
    from ...robie.childs import RobieClient
    from ...robie.models import RobieCommand
    from ...robie.models import RobieMessageKinds



@dataclass
class MTMMessage(RobieMessage):
    """
    Contain information for sharing using the Python queue.
    """

    event: ClientEvent


    def __init__(
        self,
        client: 'RobieClient',
        event: ClientEvent,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        robie = client.robie

        self.event = event

        person = (
            robie.person(
                client,
                event.author[1])
            if event.author
            else None)

        super().__init__(
            client=client,
            person=person)


    @property
    def family(
        self,
    ) -> Literal['mattermost']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'mattermost'


    @property
    def kind(
        self,
    ) -> 'RobieMessageKinds':
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.event.kind


    @property
    def isme(
        self,
    ) -> bool:
        """
        Return the boolean indicating message origin from client.

        :returns: Boolean indicating message origin from client.
        """

        return self.event.isme


    @property
    def hasme(
        self,
    ) -> bool:
        """
        Return the boolean indicating message mention the client.

        :returns: Boolean indicating message mention the client.
        """

        return self.event.hasme


    @property
    def whome(
        self,
    ) -> tuple[str, str] | None:
        """
        Return the current nickname of the client on the server.

        :returns: Current nickname of the client on the server.
        """

        return self.event.whome


    @property
    def author(
        self,
    ) -> tuple[str, str] | None:
        """
        Return the current nickname of the client on the server.

        :returns: Current nickname of the client on the server.
        """

        return self.event.author


    @property
    def anchor(
        self,
    ) -> str | None:
        """
        Return the unique value for the context with the client.

        :returns: Unique value for the context with the client.
        """

        kind = self.kind
        author = self.author
        event = self.event

        recipient = (
            event.recipient)

        if (recipient is None
                or author is None):
            return None

        return (
            recipient
            if kind == 'chanmsg'
            else author[1])


    @property
    def message(
        self,
    ) -> str | None:
        """
        Return the string containing the content of the message.

        :returns: String containing the content of the message.
        """

        return self.event.message


    def reply(
        self,
        robie: 'Robie',
        content: str,
    ) -> 'RobieCommand':
        """
        Compose the message and transmit using the chat client.

        :param robie: Primary class instance for Chatting Robie.
        :param content: Content that will be source of message.
        """

        childs = robie.childs
        clients = childs.clients

        client = clients[
            self.client]

        event = self.event


        target = event.recipient

        assert target is not None


        return client.compose(
            target, content)



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from re import IGNORECASE
from re import compile
from re import match as re_match
from typing import TYPE_CHECKING
from typing import Type

from encommon.times import Timer
from encommon.types import NCFalse
from encommon.types import NCNone
from encommon.types.strings import SPACED

from .params import AutoNickPluginParams
from .params import AutoNickPluginServiceParams
from ..status import StatusPlugin
from ..status import StatusPluginStates
from ...robie.childs import RobiePlugin

if TYPE_CHECKING:
    from ...clients.irc import IRCClient
    from ...clients.irc.message import IRCMessage



_IDENTIFY = compile(
    r'this nick(name)? is registered',
    IGNORECASE)

_SERVICES = dict[str, AutoNickPluginServiceParams]



class AutoNickPlugin(RobiePlugin):
    """
    Integrate with the Robie routine and perform operations.

    .. note::
       This plugin maintains configured nickname on server.
    """

    __started: bool

    __nickserv: _SERVICES

    __timer: Timer


    def __post__(
        self,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__started = False

        params = self.params


        nickserv: _SERVICES = {}

        if params.services:

            services = params.services

            for service in services:

                name = service.client

                nickserv[name] = service

        self.__nickserv = nickserv


        self.__timer = Timer(
            params.interval)

        self.__status('pending')


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        # Review the parameters


    @classmethod
    def schema(
        cls,
    ) -> Type[AutoNickPluginParams]:
        """
        Return the configuration parameters relevant for class.

        :returns: Configuration parameters relevant for class.
        """

        return AutoNickPluginParams


    @property
    def params(
        self,
    ) -> AutoNickPluginParams:
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        params = super().params

        assert isinstance(
            params,
            AutoNickPluginParams)

        return params


    def operate(
        self,
    ) -> None:
        """
        Perform the operation related to Robie service threads.
        """

        from ...clients.irc.message import IRCMessage

        assert self.thread

        thread = self.thread
        mqueue = thread.mqueue
        timer = self.__timer

        clients = (
            thread.service
            .clients.childs)


        if not self.__started:
            self.__started = True
            self.__status('normal')


        if timer.ready():
            self.__operate()


        while not mqueue.empty:

            mitem = mqueue.get()

            name = mitem.client
            family = mitem.family

            # Ignore unrelated clients
            if family != 'irc':
                continue

            assert isinstance(
                mitem, IRCMessage)

            # Ignore disabled clients
            if name not in clients:
                continue  # NOCVR

            self.__message(mitem)


    def __operate(
        self,
    ) -> None:
        """
        Perform the operation related to Robie service threads.
        """

        from ...clients import IRCClient

        assert self.thread

        thread = self.thread
        member = thread.member
        cqueue = member.cqueue
        params = self.params

        clients = (
            thread.service
            .clients.childs
            .values())

        names = params.clients

        failure: set[bool] = set()


        for client in clients:

            name = client.name

            # Ignore unrelated clients
            if name not in names:
                continue

            assert isinstance(
                client, IRCClient)

            connected = (
                client.client
                .connected)

            # Bypass when disconnected
            if connected is False:
                failure.add(True)
                return None


            should = (
                client.params
                .client.nickname)

            current = (
                client.client
                .nickname)

            if current == should:
                return None


            failure.add(True)

            rawcmd = f'NICK :{should}'

            client.put_command(
                cqueue, rawcmd)


        self.__status(
            'failure'
            if any(failure)
            else 'normal')


    def __message(  # noqa: CFQ004
        self,
        mitem: 'IRCMessage',
    ) -> None:
        """
        Process the provided message item from the Robie thread.

        :param mitem: Item containing information for operation.
        """

        from ...clients import IRCClient

        robie = self.robie
        nickserv = self.__nickserv
        childs = robie.childs

        event = mitem.event
        command = event.command

        client = (
            childs.clients
            [mitem.client])

        assert isinstance(
            client, IRCClient)


        name = client.name

        if name not in nickserv:
            return NCNone

        service = (
            nickserv[name]
            .service)


        def _requested() -> bool:

            # Ignore when not NickServ
            if command != 'NOTICE':
                return False

            assert event.prefix
            assert event.params

            author = (
                event.prefix
                .split('!', 1)[0])

            message = (
                event.params
                .split(SPACED, 1)[1]
                .lstrip(':'))

            match = re_match(
                _IDENTIFY, message)

            # Ignore when not NickServ
            if author != service:
                return NCFalse

            return match is not None


        if (_requested()
                or command == '376'):
            self.__identify(client)


    def __identify(
        self,
        client: 'IRCClient',
    ) -> None:
        """
        Identify with the network nick services using password.

        :param client: Class instance for connecting to service.
        """

        assert self.thread

        thread = self.thread
        nickserv = self.__nickserv
        member = thread.member
        cqueue = member.cqueue

        name = client.name

        if name not in nickserv:
            return NCNone

        service = (
            nickserv[name]
            .service)

        password = (
            nickserv[name]
            .password)

        rawcmd = (
            f'PRIVMSG {service} '
            f':identify {password}')

        client.put_command(
            cqueue, rawcmd)



    def __status(
        self,
        status: StatusPluginStates,
    ) -> None:
        """
        Update or insert the status of the Robie child instance.

        :param status: One of several possible value for status.
        """

        thread = self.thread
        params = self.params

        if thread is None:
            return None

        plugins = (
            thread.service
            .plugins.childs)

        if 'status' not in plugins:
            return NCNone

        plugin = plugins['status']

        assert isinstance(
            plugin, StatusPlugin)

        (plugin.update(
            unique=self.name,
            group='Management',
            title='IRC',
            icon=params.status,
            state=status))



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .config import RobieConfig
from .robie import Robie
from .service import RobieService



__all__ = [
    'Robie',
    'RobieConfig',
    'RobieService']



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path

from .test_client import MTMEVENTS
from .test_client import MTMEVENT_HUBERT_CHAN
from .test_client import MTMEVENT_HUBERT_PRIV
from .test_client import MTMEVENT_RANDOM_CHAN
from .test_client import MTMEVENT_RANDOM_PRIV



__all__ = [
    'MTMEVENTS',
    'MTMEVENT_HUBERT_CHAN',
    'MTMEVENT_HUBERT_PRIV',
    'MTMEVENT_RANDOM_CHAN',
    'MTMEVENT_RANDOM_PRIV']



SAMPLES = (
    Path(__file__).parent
    / 'samples')



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path
from threading import Thread
from time import sleep as block_sleep
from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs
from encommon.utils import read_text

from enconnect.fixtures import DSCClientSocket
from enconnect.fixtures import IRCClientSocket
from enconnect.fixtures import MTMClientSocket

from ..plugin import LoggerPlugin
from ....clients.discord.test import DSCEVENTS
from ....clients.irc.test import IRCEVENTS
from ....clients.mattermost.test import MTMEVENTS

if TYPE_CHECKING:
    from ....robie import RobieService



def test_LoggerPlugin(
    service: 'RobieService',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param service: Ancilary Chatting Robie class instance.
    """

    plugin = (
        service.plugins
        .childs['logger'])

    assert isinstance(
        plugin, LoggerPlugin)


    attrs = lattrs(plugin)

    assert attrs == [
        '_RobieChild__robie',
        '_RobieChild__name',
        '_RobieChild__params',
        '_LoggerPlugin__started',
        '_LoggerPlugin__history',
        '_RobiePlugin__thread']


    assert inrepr(
        'plugin.LoggerPlugin',
        plugin)

    assert isinstance(
        hash(plugin), int)

    assert instr(
        'plugin.LoggerPlugin',
        plugin)


    plugin.validate()

    assert plugin.robie

    assert plugin.enable

    assert plugin.name == 'logger'

    assert plugin.kind == 'plugin'

    assert plugin.schema()

    assert plugin.params

    assert plugin.thread

    assert plugin.dumped



def test_LoggerPlugin_cover(
    service: 'RobieService',
    client_dscsock: DSCClientSocket,
    client_ircsock: IRCClientSocket,
    client_mtmsock: MTMClientSocket,
    tmp_path: Path,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param service: Ancilary Chatting Robie class instance.
    :param client_dscsock: Object to mock client connection.
    :param client_ircsock: Object to mock client connection.
    :param client_mtmsock: Object to mock client connection.
    :param tmp_path: pytest object for temporal filesystem.
    """

    plugin = (
        service.plugins
        .childs['logger'])

    assert isinstance(
        plugin, LoggerPlugin)

    plugin.params.output = (
        f'{tmp_path}/output.txt')


    client_dscsock(DSCEVENTS)
    client_ircsock(IRCEVENTS)
    client_mtmsock(MTMEVENTS)


    service.limit(
        plugins=[
            'logger',
            'status'])

    service.start()


    thread = Thread(
        target=service.operate)

    thread.start()


    block_sleep(5)

    service.soft()

    while service.running:
        block_sleep(1)

    service.stop()

    thread.join()


    content = read_text(
        f'{tmp_path}/output.txt')

    assert 'Hello dscbot' in content
    assert 'Hello ircbot' in content
    assert 'Hello mtmbot' in content



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



import asyncio

from encommon.types import funcname

from pydantic_ai import RunContext

from .persist import HomiePersistRecord
from ..ainswer import AinswerDepends



async def homie_persist(
    context: RunContext[AinswerDepends],
) -> list[HomiePersistRecord]:
    """
    Return the current values in Homie Automate persistence.

    .. note::
       This tool will return values from the Homie Automate
       persistence store. They may include information about
       many things regarding the Robie owner home status.

    :returns: Current values in Homie Automate persistence.
    """

    from .plugin import HomiePlugin

    deps = context.deps
    ainswer = deps.plugin
    mitem = deps.mitem

    assert ainswer.thread

    assert mitem is not None

    thread = ainswer.thread

    plugins = (
        thread.service
        .plugins.childs
        .values())


    returned: list[HomiePersistRecord] = []


    for plugin in plugins:

        related = isinstance(
            plugin, HomiePlugin)

        if related is False:
            continue

        assert isinstance(
            plugin, HomiePlugin)

        # Basic trust enforcement
        if plugin.notrust(mitem):
            continue

        returned.extend(
            plugin.persist
            .records)


    await asyncio.sleep(0)

    ainswer.printer({
        'Function': funcname(),
        'Returned': returned})

    return returned



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Literal
from typing import TYPE_CHECKING

from encommon.times import Time
from encommon.types import BaseModel

from enconnect.utils import HTTPClient

from httpx import Response

from pydantic import Field

if TYPE_CHECKING:
    from .plugin import NagiosPlugin



_FIELDS = {
    'status': 'last_hard_state',
    'description': 'description',
    'downtime': 'scheduled_downtime_depth',
    'handled': 'problem_has_been_acknowledged',
    'checked': 'last_check'}



class NagiosCurrentObject(BaseModel, extra='forbid'):
    """
    Information regarding the item within Nagios monitoring.
    """

    name: Annotated[
        str,
        Field(...,
              description='Name of the object in Nagios',
              min_length=1)]

    status: Annotated[
        str,
        Field(...,
              description='Current status for the object',
              min_length=1)]

    latest: Annotated[
        str,
        Field(...,
              description='When the latest event occurred',
              min_length=1)]

    downtime: Annotated[
        bool,
        Field(...,
              description='Whether in scheduled downtime')]

    handled: Annotated[
        bool,
        Field(...,
              description='Whether or not if was handled')]



class NagiosCurrentHost(NagiosCurrentObject, extra='forbid'):
    """
    Information regarding the item within Nagios monitoring.
    """



class NagiosCurrentService(NagiosCurrentObject, extra='forbid'):
    """
    Information regarding the item within Nagios monitoring.
    """

    host: Annotated[
        str,
        Field(...,
              description='Host the service is related to',
              min_length=1)]



class NagiosCurrentRecords(BaseModel, extra='forbid'):
    """
    Summarized information regarding the items within Nagios.
    """

    hosts: Annotated[
        list[NagiosCurrentHost],
        Field(...,
              description='System related status values')]

    services: Annotated[
        list[NagiosCurrentService],
        Field(...,
              description='Service related status values')]



class NagiosCurrentSummary(BaseModel, extra='forbid'):
    """
    Summarized information regarding the items within Nagios.
    """

    service_normal: Annotated[
        int,
        Field(...,
              description='Object count in normal state',
              ge=0)]

    service_issues: Annotated[
        int,
        Field(...,
              description='Object count in issue state',
              ge=0)]

    host_normal: Annotated[
        int,
        Field(...,
              description='Object count in normal state',
              ge=0)]

    host_issues: Annotated[
        int,
        Field(...,
              description='Object count in issue state',
              ge=0)]



class NagiosCurrent:
    """
    Collect the information using the upstream API endpoint.

    :param plugin: Plugin class instance for Chatting Robie.
    """

    __plugin: 'NagiosPlugin'


    def __init__(
        self,
        plugin: 'NagiosPlugin',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__plugin = plugin


    def request(
        self,
        item: Literal['host', 'service'],
    ) -> Response:
        """
        Return the response from upstream API endpoint request.

        :param item: Determine which items will be enumerated.
        :returns: Response from upstream API endpoint request.
        """

        plugin = self.__plugin
        params = plugin.params


        client = HTTPClient(
            timeout=params.timeout,
            verify=params.ssl_verify,
            capem=params.ssl_capem)

        request = client.request_block

        response = request(
            method='get',
            location=(
                f'{params.restful}'
                '/cgi-bin/statusjson.cgi'),
            httpauth=(
                (params.username,
                 params.password)
                if params.username
                and params.password
                else None),
            timeout=params.timeout,
            params={
                'query': f'{item}list',
                'formatoptions': 'enumerate',
                'details': 'true'})

        (response
         .raise_for_status())


        return response


    @property
    def hosts(
        self,
    ) -> list[NagiosCurrentHost]:
        """
        Return the response from upstream API endpoint request.

        :returns: Response from upstream API endpoint request.
        """

        returned: list[NagiosCurrentHost] = []

        model = NagiosCurrentHost


        response = self.request('host')


        items = (
            response.json()
            ['data']['hostlist']
            .items())

        for name, values in items:

            checked = values[
                _FIELDS['checked']]

            latest = Time(
                checked / 1000,
                tzname='US/Central')

            status = values[
                _FIELDS['status']]

            downtime = values[
                _FIELDS['downtime']]

            handled = values[
                _FIELDS['handled']]

            object = model(
                name=name,
                status=status,
                latest=latest.human,
                downtime=downtime,
                handled=handled)

            returned.append(object)


        return returned


    @property
    def services(
        self,
    ) -> list[NagiosCurrentService]:
        """
        Return the response from upstream API endpoint request.

        :returns: Response from upstream API endpoint request.
        """

        returned: list[NagiosCurrentService] = []

        model = NagiosCurrentService


        response = self.request('service')


        items = (
            response.json()
            ['data']['servicelist']
            .items())


        for host, _values in items:

            _items = _values.values()

            for values in _items:

                checked = values[
                    _FIELDS['checked']]

                latest = Time(
                    checked / 1000,
                    tzname='US/Central')

                name = values[
                    _FIELDS['description']]

                status = values[
                    _FIELDS['status']]

                downtime = values[
                    _FIELDS['downtime']]

                handled = values[
                    _FIELDS['handled']]

                object = model(
                    name=name,
                    host=host,
                    status=status,
                    latest=latest.human,
                    downtime=downtime,
                    handled=handled)

                returned.append(object)


        return returned


    @property
    def summary(
        self,
    ) -> NagiosCurrentSummary:
        """
        Return the response from upstream API endpoint request.

        :returns: Response from upstream API endpoint request.
        """

        normal = {'host': 0, 'service': 0}
        issues = {'host': 0, 'service': 0}

        for host in self.hosts:

            target = (
                normal
                if host.status == 'up'
                else issues)

            target['host'] += 1

        for srvc in self.services:

            target = (
                normal
                if srvc.status == 'ok'
                else issues)

            target['service'] += 1

        return NagiosCurrentSummary(
            service_normal=normal['service'],
            service_issues=issues['service'],
            host_normal=normal['host'],
            host_issues=issues['host'])


    @property
    def records(
        self,
    ) -> NagiosCurrentRecords:
        """
        Return the response from upstream API endpoint request.

        :returns: Response from upstream API endpoint request.
        """

        return NagiosCurrentRecords(
            services=self.services,
            hosts=self.hosts)



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from json import dumps
from typing import TYPE_CHECKING
from typing import Type

from encommon.types import NCNone
from encommon.utils import append_text

from .history import LoggerHistory
from .params import LoggerPluginParams
from ..status import StatusPlugin
from ..status import StatusPluginStates
from ...robie.childs import RobiePlugin

if TYPE_CHECKING:
    from ...robie.models import RobieMessage



class LoggerPlugin(RobiePlugin):
    """
    Integrate with the Robie routine and perform operations.

    .. note::
       This plugin allows for interacting with an LLM model.
    """

    __started: bool

    __history: LoggerHistory


    def __post__(
        self,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__started = False

        self.__history = (
            LoggerHistory(self))

        self.__status('pending')


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        # Review the parameters


    @classmethod
    def schema(
        cls,
    ) -> Type[LoggerPluginParams]:
        """
        Return the configuration parameters relevant for class.

        :returns: Configuration parameters relevant for class.
        """

        return LoggerPluginParams


    @property
    def params(
        self,
    ) -> LoggerPluginParams:
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        params = super().params

        assert isinstance(
            params,
            LoggerPluginParams)

        return params


    @property
    def history(
        self,
    ) -> LoggerHistory:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__history


    def operate(
        self,
    ) -> None:
        """
        Perform the operation related to Robie service threads.
        """

        assert self.thread

        thread = self.thread
        mqueue = thread.mqueue
        params = self.params
        history = self.history

        clients = (
            thread.service
            .clients.childs)

        names = params.clients
        output = params.output


        if not self.__started:
            self.__started = True
            self.__status('normal')


        kinds = ['privmsg', 'chanmsg']

        while not mqueue.empty:

            mitem = mqueue.get()

            name = mitem.client
            kind = mitem.kind

            # Ignore unrelated clients
            if name not in names:
                continue  # NOCVR

            # Ignore unrelated events
            if kind not in kinds:
                continue

            # Ignore disabled clients
            if name not in clients:
                continue  # NOCVR


            history.process(mitem)

            self.__status('normal')


            if output is not None:
                self.__writeout(mitem)


    def __writeout(
        self,
        mitem: 'RobieMessage',
    ) -> None:
        """
        Write the event out to the designated output file path.

        :param mitem: Item containing information for operation.
        """

        robie = self.robie
        childs = robie.childs
        persons = childs.persons
        clients = childs.clients
        params = self.params

        output = params.output

        _client = mitem.client
        _person = mitem.person
        author = mitem.author
        anchor = mitem.anchor
        message = mitem.message

        person = (
            persons[_person]
            if _person is not None
            else None)

        client = clients[_client]

        assert output is not None
        assert author is not None
        assert anchor is not None
        assert message is not None


        try:

            append = {
                'time': str(mitem.time),
                'client': client.name,
                'person': (
                    person.name
                    if person is not None
                    else None),
                'author': author,
                'anchor': anchor,
                'message': message}

            append_text(
                output,
                f'{dumps(append)}\n')

            self.__status('normal')

        except Exception as reason:

            self.__status('failure')

            robie.logger.log_e(
                base=self,
                name=self,
                item='writeout',
                status='exception',
                exc_info=reason)



    def __status(
        self,
        status: StatusPluginStates,
    ) -> None:
        """
        Update or insert the status of the Robie child instance.

        :param status: One of several possible value for status.
        """

        thread = self.thread
        params = self.params

        if thread is None:
            return None

        plugins = (
            thread.service
            .plugins.childs)

        if 'status' not in plugins:
            return NCNone

        plugin = plugins['status']

        assert isinstance(
            plugin, StatusPlugin)

        (plugin.update(
            unique=self.name,
            group='Extensions',
            title='Logger',
            icon=params.status,
            state=status))



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .child import RobieChild
from .client import RobieClient
from .person import RobiePerson
from .plugin import RobiePlugin
from .robie import RobieChilds



__all__ = [
    'RobieChild',
    'RobieChilds',
    'RobieClient',
    'RobiePlugin',
    'RobiePerson']



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from ..robie import RobieModels



def test_RobieModels_cover() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    models = RobieModels

    assert models.robie()
    assert models.printer()
    assert models.service()

    assert models.child()
    assert models.client()
    assert models.plugin()
    assert models.person()

    assert models.queue()
    assert models.message()
    assert models.command()



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Any
from typing import Callable
from typing import Optional
from typing import TYPE_CHECKING

from encommon.types import BaseModel
from encommon.types import NCNone

from pydantic import Field

if TYPE_CHECKING:
    from ..robie.models import RobieMessage



class ClientChannel(BaseModel, extra='forbid'):
    """
    Cache the information regarding client status on server.
    """

    unique: Annotated[
        str,
        Field(...,
              description='Unique ID of channel on server',
              min_length=1)]

    title: Annotated[
        Optional[str],
        Field(None,
              description='Friendly human name of channel',
              min_length=1)]

    topic: Annotated[
        Optional[str],
        Field(None,
              description='Current topic for the channel',
              min_length=1)]

    members: Annotated[
        Optional[set[str]],
        Field(None,
              description='Current members of the channel',
              min_length=1)]


    def __init__(
        # NOCVR
        self,
        /,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        super().__init__(**data)



class ClientChannels:
    """
    Cache the information regarding client status on server.

    :param client: Client class instance for Chatting Robie.
    """

    __cached: dict[str, ClientChannel]


    def __init__(
        self,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__cached = {}


    def create(
        self,
        unique: str,
    ) -> None:
        """
        Create the channel within the internal cache dictionary.

        :param unique: Unique identifier for channel on server.
        """

        cached = self.__cached

        item = ClientChannel(
            unique=unique)

        cached[unique] = item


    def select(
        self,
        unique: str,
    ) -> Optional[ClientChannel]:
        """
        Select the channel within the internal cache dictionary.

        :param unique: Unique identifier for channel on server.
        :returns: Channel within the internal cache dictionary.
        """

        cached = self.__cached

        names = {
            x.title: x.unique
            for x in
            cached.values()}

        if unique not in cached:

            if unique not in names:
                return None

            unique = names[unique]

        return cached[unique]


    def delete(
        self,
        unique: str,
    ) -> None:
        """
        Delete the channel within the internal cache dictionary.

        :param unique: Unique identifier for channel on server.
        """

        cached = self.__cached

        del cached[unique]


    def set_title(
        self,
        unique: str,
        title: str,
    ) -> None:
        """
        Update attribute for channel and create when not exists.

        :param unique: Unique identifier for channel on server.
        :param title: Proper huamn friendly name of the channel.
        """

        cached = self.__cached

        if unique not in cached:
            self.create(unique)

        select = self.select(unique)

        assert select is not None

        select.title = title or None


    def set_topic(
        self,
        unique: str,
        topic: str,
    ) -> None:
        """
        Update attribute for channel and create when not exists.

        :param unique: Unique identifier for channel on server.
        :param topic: Topic for discussion within the channel.
        """

        cached = self.__cached

        if unique not in cached:
            self.create(unique)

        select = self.select(unique)

        assert select is not None

        select.topic = topic or None


    def add_member(
        self,
        unique: str,
        member: str,
    ) -> None:
        """
        Update attribute for channel and create when not exists.

        :param unique: Unique identifier for channel on server.
        :param member: Remember relevant to the cache operation.
        """

        cached = self.__cached

        if unique not in cached:
            self.create(unique)

        select = self.select(unique)

        assert select is not None

        if not select.members:
            select.members = set()

        members = select.members

        members.add(member)


    def del_member(
        self,
        unique: str,
        member: str,
    ) -> None:
        """
        Update attribute for channel and create when not exists.

        :param unique: Unique identifier for channel on server.
        :param member: Remember relevant to the cache operation.
        """

        cached = self.__cached

        if unique not in cached:
            return NCNone

        select = self.select(unique)

        assert select is not None

        if not select.members:
            return NCNone

        members = select.members

        members.remove(member)

        if len(members) == 0:
            select.members = None


    def clear_members(
        self,
        unique: str,
    ) -> None:
        """
        Update attribute for channel and create when not exists.

        :param unique: Unique identifier for channel on server.
        """

        cached = self.__cached

        if unique not in cached:
            return NCNone

        select = self.select(unique)

        assert select is not None

        if not select.members:
            return NCNone

        select.members = None


    def rename_member(
        self,
        current: str,
        update: str,
    ) -> None:
        """
        Update attribute for channel and create when not exists.

        :param current: Current member name to be be searched.
        :param update: Value to replace current when is found.
        """

        cached = (
            self.__cached
            .values())

        for channel in cached:

            members = channel.members

            if members is None:
                continue

            if current not in members:
                continue

            members.remove(current)
            members.add(update)



class ClientPublish:
    """
    Allow for subscription to client events useful in tests.
    """

    __callbacks: list[Callable[['RobieMessage'], None]]


    def __init__(
        self,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__callbacks = []


    def subscribe(
        self,
        callback: Callable[['RobieMessage'], None],
    ) -> None:
        """
        Store the callback within internal subscriber reference.

        :param callback: Function that will be called with item.
        """

        callbacks = self.__callbacks

        callbacks.append(callback)


    def publish(
        self,
        mitem: 'RobieMessage',
    ) -> None:
        """
        Submit the message from client to subscriber references.

        :param mitem: Item containing information for operation.
        """

        callbacks = self.__callbacks

        for callback in callbacks:
            callback(mitem)



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from threading import Lock
from typing import Annotated
from typing import Any
from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING

from encommon.times import Time
from encommon.types import BaseModel
from encommon.types import DictStrAny
from encommon.types import NCNone

from pydantic import Field

from sqlalchemy import Column
from sqlalchemy import Float
from sqlalchemy import String
from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import Session
from sqlalchemy.orm import sessionmaker

if TYPE_CHECKING:
    from .plugin import AinswerPlugin
    from ...robie.models import RobieMessage



AinswerHistoryKinds = Literal[
    'privmsg', 'chanmsg']



class SQLBase(DeclarativeBase):
    """
    Some additional class that SQLAlchemy requires to work.
    """



class AinswerHistoryTable(SQLBase):
    """
    Schematic for the database operations using SQLAlchemy.

    .. note::
       Fields are not completely documented for this model.
    """

    plugin = Column(
        String,
        primary_key=True,
        nullable=False)

    client = Column(
        String,
        primary_key=True,
        nullable=False)

    person = Column(
        String,
        primary_key=True,
        nullable=True)

    kind = Column(
        String,
        primary_key=True,
        nullable=False)

    author = Column(
        String,
        primary_key=True,
        nullable=False)

    anchor = Column(
        String,
        primary_key=True,
        nullable=False)

    message = Column(
        String,
        nullable=False)

    ainswer = Column(
        String,
        nullable=False)

    create = Column(
        Float,
        primary_key=True,
        nullable=False)

    __tablename__ = 'ainswer'



class AinswerHistoryRecord(BaseModel, extra='forbid'):
    """
    Contain the information regarding the chatting history.

    :param record: Record from the SQLAlchemy query routine.
    :param kwargs: Keyword arguments passed for downstream.
    """

    plugin: Annotated[
        str,
        Field(...,
              description='Plugin name where originated',
              min_length=1)]

    client: Annotated[
        str,
        Field(...,
              description='Client name where originated',
              min_length=1)]

    person: Annotated[
        Optional[str],
        Field(None,
              description='Person that author is matched',
              min_length=1)]

    kind: Annotated[
        AinswerHistoryKinds,
        Field(...,
              description='What kind of original message',
              min_length=1)]

    author: Annotated[
        str,
        Field(...,
              description='Nickname of message author',
              min_length=1)]

    anchor: Annotated[
        str,
        Field(None,
              description='Channel or private context',
              min_length=1)]

    message: Annotated[
        str,
        Field(...,
              description='Historical interaction content')]

    ainswer: Annotated[
        str,
        Field(...,
              description='Historical interaction content')]

    create: Annotated[
        str,
        Field(...,
              description='When the record was created',
              min_length=20,
              max_length=32)]


    def __init__(
        self,
        record: Optional[AinswerHistoryTable] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        params: DictStrAny

        fields = [
            'plugin',
            'client',
            'person',
            'kind',
            'author',
            'anchor',
            'message',
            'ainswer',
            'create']


        if record is not None:

            params = {
                x: getattr(record, x)
                for x in fields}


        elif record is None:

            params = {
                x: kwargs.get(x)
                for x in fields}


        else:  # NOCVR
            raise ValueError('record')


        create = params['create']

        params['create'] = (
            str(Time(create)))


        super().__init__(**params)



class AinswerHistory:
    """
    Store the historical information for chat interactions.

    :param plugin: Plugin class instance for Chatting Robie.
    """

    __plugin: 'AinswerPlugin'

    __connect: str
    __locker: Lock

    __sengine: Engine
    __session: (
        # pylint: disable=unsubscriptable-object
        sessionmaker[Session])


    def __init__(
        self,
        plugin: 'AinswerPlugin',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__plugin = plugin

        params = plugin.params

        self.__connect = (
            params.database)

        self.__locker = Lock()

        self.__build_engine()


    def __build_engine(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        path = self.__connect

        sengine = (
            create_engine(path))

        (SQLBase.metadata
         .create_all(sengine))

        session = (
            sessionmaker(sengine))

        self.__sengine = sengine
        self.__session = session


    def insert(  # noqa: CFQ002
        self,
        *,
        client: str,
        person: str | None,
        kind: str,
        author: str,
        anchor: str,
        message: str,
        ainswer: str,
    ) -> None:
        """
        Insert the record into the historical chat interactions.
        """

        plugin = self.__plugin

        sess = self.__session()
        lock = self.__locker

        table = AinswerHistoryTable
        model = AinswerHistoryRecord


        inputs: DictStrAny = {
            'plugin': plugin.name,
            'client': client,
            'person': person,
            'kind': kind,
            'author': author,
            'anchor': anchor,
            'message': message,
            'ainswer': ainswer,
            'create': float(Time())}

        record = model(**inputs)

        insert = record.endumped


        create = insert['create']

        insert['create'] = (
            float(Time(create)))


        with lock, sess as session:

            session.merge(
                table(**insert))

            session.commit()


        self.expunge(
            client=record.client,
            kind=record.kind,
            anchor=record.anchor)


    def process(
        self,
        mitem: 'RobieMessage',
        ainswer: str,
    ) -> None:
        """
        Insert the record into the historical chat interactions.

        :param mitem: Item containing information for operation.
        :param ainswer: Response from LLM relevant to question.
        """

        plugin = self.__plugin
        robie = plugin.robie
        childs = robie.childs
        persons = childs.persons
        clients = childs.clients

        assert mitem.author
        assert mitem.anchor
        assert mitem.message

        _person = mitem.person
        _kind = mitem.kind
        _author = mitem.author
        _anchor = mitem.anchor
        _message = mitem.message

        client = clients[
            mitem.client]

        person = (
            persons[_person]
            if _person is not None
            else None)

        self.insert(
            client=client.name,
            person=(
                person.name
                if person is not None
                else None),
            kind=_kind,
            author=_author[1],
            anchor=_anchor,
            message=_message,
            ainswer=ainswer)


    def expunge(
        self,
        *,
        client: Optional[str] = None,
        kind: Optional[str] = None,
        anchor: Optional[str] = None,
    ) -> None:
        """
        Remove the expired historical chat interaction records.
        """

        plugin = self.__plugin
        params = plugin.params

        maximum = params.histories


        sess = self.__session()
        lock = self.__locker

        table = AinswerHistoryTable

        _plugin = table.plugin
        _client = table.client
        _kind = table.kind
        _anchor = table.anchor
        _create = table.create


        with lock, sess as session:

            total = (
                session.query(table)
                .filter(
                    _plugin == plugin.name,
                    _client == client,
                    _kind == kind,
                    _anchor == anchor)
                .count())

            if total <= maximum:
                return

            cutoff = (
                session.query(_create)
                .filter(
                    _plugin == plugin.name,
                    _client == client,
                    _kind == kind,
                    _anchor == anchor)
                .order_by(_create.desc())
                .offset(maximum - 1)
                .limit(1).scalar())

            if cutoff is None:
                return NCNone

            (session.query(table)
             .filter(
                 _plugin == plugin.name,
                 _client == client,
                 _kind == kind,
                 _anchor == anchor,
                 _create < cutoff)
             .delete(synchronize_session=False))

            session.commit()


    def records(
        self,
        mitem: 'RobieMessage',
        limit: Optional[int] = None,
    ) -> list[AinswerHistoryRecord]:
        """
        Return all historical records for the chat interactions.

        :param mitem: Item containing information for operation.
        :param limit: Optionally restrict the records returned.
        :returns: Historical records for the chat interactions.
        """

        plugin = self.__plugin
        robie = plugin.robie
        childs = robie.childs
        clients = childs.clients

        assert mitem.author

        client = clients[
            mitem.client]


        return self.search(
            limit=limit,
            client=client.name,
            kind=mitem.kind,
            anchor=mitem.anchor)


    def search(  # noqa: CFQ002
        self,
        limit: Optional[int] = None,
        *,
        client: Optional[str] = None,
        person: Optional[str] = None,
        kind: Optional[str] = None,
        author: Optional[str] = None,
        anchor: Optional[str] = None,
    ) -> list[AinswerHistoryRecord]:
        """
        Return all historical records for the chat interactions.

        :param limit: Optionally restrict the records returned.
        :returns: Historical records for the chat interactions.
        """

        plugin = self.__plugin


        sess = self.__session()
        lock = self.__locker

        records: list[AinswerHistoryRecord]

        table = AinswerHistoryTable
        model = AinswerHistoryRecord

        _plugin = table.plugin
        _client = table.client
        _person = table.person
        _kind = table.kind
        _author = table.author
        _anchor = table.anchor
        _create = table.create


        with lock, sess as session:

            records = []

            query = (
                session.query(table)
                .filter(
                    _plugin == plugin.name)
                .order_by(_create.desc()))

            if client is not None:
                query = query.filter(
                    _client == client)

            if person is not NCNone:
                query = query.filter(
                    _person == person)

            if kind is not None:
                query = query.filter(
                    _kind == kind)

            if author is not None:
                query = query.filter(
                    _author == author)

            if anchor is not None:
                query = query.filter(
                    _anchor == anchor)

            if limit is not None:
                query = query.limit(limit)


            for record in query.all():

                object = model(record)

                records.append(object)

            return records[::-1]



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from threading import Event
from typing import TYPE_CHECKING

from encommon.types import clsname

from ..addons import RobieQueue
from ..threads import RobieThread  # noqa: F401

if TYPE_CHECKING:
    from ..models import RobieCommand
    from ..models import RobieMessage
    from ..robie import Robie
    from ..service import RobieService



RobieThreads = dict[str, 'RobieThread']

_CONGEST = list[tuple[str, int]]



class RobieMember:
    """
    Common methods and routines for Chatting Robie members.

    :param robie: Primary class instance for Chatting Robie.
    """

    __service: 'RobieService'

    __threads: RobieThreads

    __mqueue: RobieQueue['RobieMessage']
    __cqueue: RobieQueue['RobieCommand']
    __vacate: Event
    __cancel: Event


    def __init__(
        self,
        service: 'RobieService',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        robie = service.robie

        robie.logger.log_d(
            base=self,
            status='initial')

        self.__service = service

        self.__threads = {}

        self.__build_objects()

        robie.logger.log_i(
            base=self,
            status='created')


    def __build_objects(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        robie = self.robie

        self.__mqueue = (
            RobieQueue(robie))

        self.__cqueue = (
            RobieQueue(robie))

        self.__vacate = Event()
        self.__cancel = Event()

        self.build_threads()


    def build_threads(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.

        .. note::
           Deviates from enhomie in build happens downstream,
           because all of the enhomie members are origin based
           where enrobie children do not have consistent base.
        """

        raise NotImplementedError


    def limit(
        self,
        names: list[str],
    ) -> None:
        """
        Remove the thread from member when not already started.

        .. note::
           Deviates from enhomie in build happens downstream,
           because all of the enhomie members are origin based
           where enrobie children do not have consistent base.

        :param names: Names of the children that are permitted.
        """

        threads = self.__threads

        for name in list(threads):

            thread = threads[name]

            if name in names:
                continue

            assert not thread.ident

            del threads[name]


    @property
    def robie(
        self,
    ) -> 'Robie':
        """
        Return the Robie instance to which the instance belongs.

        :returns: Robie instance to which the instance belongs.
        """

        return self.__service.robie


    @property
    def service(
        self,
    ) -> 'RobieService':
        """
        Return the Robie instance to which the instance belongs.

        :returns: Robie instance to which the instance belongs.
        """

        return self.__service


    @property
    def threads(
        self,
    ) -> RobieThreads:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        threads = self.__threads

        return dict(threads)


    @threads.setter
    def threads(
        self,
        value: RobieThreads,
    ) -> None:
        """
        Update the value for the attribute from class instance.
        """

        self.__threads = value


    @property
    def mqueue(
        self,
    ) -> RobieQueue['RobieMessage']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__mqueue


    @property
    def cqueue(
        self,
    ) -> RobieQueue['RobieCommand']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__cqueue


    @property
    def vacate(
        self,
    ) -> Event:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__vacate


    @property
    def cancel(
        self,
    ) -> Event:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__cancel


    @property
    def running(
        self,
    ) -> list[str]:
        """
        Return the list of threads which are determined running.

        :returns: List of threads which are determined running.
        """

        running: list[str] = []

        threads = (
            self.__threads
            .values())

        for thread in threads:

            if not thread.is_alive():
                continue

            name = thread.name

            running.append(name)

        return sorted(running)


    @property
    def zombies(
        self,
    ) -> list[str]:
        """
        Return the list of threads which are determined zombies.

        :returns: List of threads which are determined zombies.
        """

        zombies: list[str] = []

        threads = (
            self.__threads
            .values())

        for thread in threads:

            if thread.is_alive():
                continue

            name = thread.name

            zombies.append(name)

        return sorted(zombies)


    @property
    def congest(
        self,
    ) -> _CONGEST:
        """
        Return the list of congested threads and members queues.

        :returns: List of congested threads and members queues.
        """

        congest: _CONGEST = []

        name = clsname(self)

        mqueue = self.mqueue
        cqueue = self.cqueue


        if mqueue.qsize > 5:

            append = (
                f'{name}/mqueue',
                mqueue.qsize)

            congest.append(append)

        if cqueue.qsize > 5:

            append = (
                f'{name}/cqueue',
                cqueue.qsize)

            congest.append(append)


        threads = (
            self.__threads
            .values())

        for thread in threads:

            congest.extend(
                thread.congest)


        return sorted(congest)


    @property
    def enqueue(
        self,
    ) -> _CONGEST:
        """
        Return the list of congested threads and members queues.

        :returns: List of congested threads and members queues.
        """

        enqueue: _CONGEST = []

        name = clsname(self)

        mqueue = self.mqueue
        cqueue = self.cqueue


        if mqueue.qsize > 0:

            append = (
                f'{name}/mqueue',
                mqueue.qsize)

            enqueue.append(append)

        if cqueue.qsize > 0:

            append = (
                f'{name}/cqueue',
                cqueue.qsize)

            enqueue.append(append)


        threads = (
            self.__threads
            .values())

        for thread in threads:

            enqueue.extend(
                thread.enqueue)


        return sorted(enqueue)


    def start(
        self,
    ) -> None:
        """
        Start the various threads within the Robie class object.
        """

        threads = (
            self.__threads
            .values())

        for thread in threads:
            thread.start()


    def operate(
        self,
    ) -> None:
        """
        Perform the operation related to Robie service members.
        """

        raise NotImplementedError


    def soft(
        self,
    ) -> None:
        """
        Stop the various threads within the Robie class object.
        """

        self.__vacate.set()


    def stop(
        self,
    ) -> None:
        """
        Stop the various threads within the Robie class object.
        """

        self.__vacate.set()
        self.__cancel.set()

        threads = (
            self.__threads
            .values())

        for thread in threads:
            thread.stop()



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path
from threading import Thread
from time import sleep as block_sleep
from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs
from encommon.utils import load_sample
from encommon.utils import prep_sample
from encommon.utils.sample import ENPYRWS

from enconnect.fixtures import MTMClientSocket
from enconnect.mattermost import ClientEvent

from pytest import raises

from . import SAMPLES
from .test_client import MTMEVENTS
from .test_client import MTMEVENT_HUBERT_CHAN
from .test_client import MTMEVENT_RANDOM_CHAN
from .test_client import MTMEVENT_RANDOM_PRIV
from ..client import MTMClient
from ..command import MTMCommand
from ..message import MTMMessage

if TYPE_CHECKING:
    from ....robie import Robie
    from ....robie import RobieService
    from ....robie.models import RobieMessage



def test_MTMMessage(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    clients = childs.clients

    client = clients['mtmbot']

    assert isinstance(
        client, MTMClient)


    item = MTMMessage(
        client,
        ClientEvent(
            client.client,
            MTMEVENT_HUBERT_CHAN))


    attrs = lattrs(item)

    assert attrs == [
        'event',
        'client',
        'person',
        'time']


    assert inrepr(
        'MTMMessage',
        item)

    with raises(TypeError):
        hash(item)

    assert instr(
        'MTMMessage',
        item)


    assert item.time.since > 0

    assert item.client == client.name

    assert item.family == 'mattermost'

    assert item.kind == 'chanmsg'

    assert item.person == 'hubert'

    assert not item.isme

    assert not item.hasme

    assert not item.whome

    assert item.author
    assert item.author[0] == 'hubert'

    assert item.anchor == 'enrobie'

    assert item.message


    event = item.event

    assert event.type == 'posted'
    assert event.data
    assert len(event.data) == 3
    assert event.broadcast
    assert len(event.broadcast) == 1
    assert event.seqno == 6
    assert not event.status
    assert not event.error
    assert not event.seqre

    assert event.kind == 'chanmsg'
    assert not event.isme
    assert not event.hasme
    assert not event.whome
    assert event.author == (
        'hubert', 'kjf9al2klaiietalkw')
    assert event.recipient == 'enrobie'
    assert event.message == 'mtmbot'



def test_MTMMessage_reply(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    clients = childs.clients

    client = clients['mtmbot']

    assert isinstance(
        client, MTMClient)


    item = MTMMessage(
        client,
        ClientEvent(
            client.client,
            MTMEVENT_RANDOM_CHAN))

    reply = item.reply(
        robie, 'Hello')

    assert isinstance(
        reply, MTMCommand)

    assert reply.json == {
        'channel_id': 'enrobie',
        'message': 'Hello'}


    item = MTMMessage(
        client,
        ClientEvent(
            client.client,
            MTMEVENT_RANDOM_PRIV))

    reply = item.reply(
        robie, 'Hello')

    assert isinstance(
        reply, MTMCommand)

    assert reply.json == {
        'channel_id': 'privid',
        'message': 'Hello'}



def test_MTMMessage_samples(
    tmp_path: Path,
    service: 'RobieService',
    client_mtmsock: MTMClientSocket,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param tmp_path: pytest object for temporal filesystem.
    :param service: Ancilary Chatting Robie class instance.
    :param client_mtmsock: Object to mock client connection.
    """

    robie = service.robie
    childs = robie.childs
    clients = childs.clients

    client = clients['mtmbot']

    assert isinstance(
        client, MTMClient)


    messages: list[MTMMessage] = []


    def _callback(
        mitem: 'RobieMessage',
    ) -> None:

        assert isinstance(
            mitem, MTMMessage)

        messages.append(mitem)


    (client.publish
     .subscribe(_callback))


    client_mtmsock(MTMEVENTS)

    service.limit(
        clients=['mtmbot'],
        plugins=['status'])

    service.start()


    thread = Thread(
        target=service.operate)

    thread.start()


    block_sleep(5)


    sample_path = (
        SAMPLES / 'messages.json')

    # Because of ones in helper
    maximum = len(MTMEVENTS) + 5

    _messages = [
        {'client': x.client,
         'family': x.family,
         'kind': x.kind,
         'person': x.person,
         'isme': x.isme,
         'hasme': x.hasme,
         'whome': x.whome,
         'author': x.author,
         'anchor': x.anchor,
         'message': x.message,
         'event': x.event}
        for x in
        messages[:maximum]]

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=_messages)

    expect = prep_sample(
        content=_messages)

    assert expect == sample


    service.soft()

    while service.running:
        block_sleep(1)

    service.stop()

    thread.join()



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from importlib import import_module
from typing import Any



def importer(
    path: str,
) -> Any:  # noqa: ANN401
    """
    Return the function or class using the importable path.

    :param path: Importable Python path for the operation.
    :returns: Function or class using the importable path.
    """

    root, name = (
        path.rsplit('.', 1))

    module = import_module(root)

    return getattr(module, name)



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

from ..states import ClientChannels
from ..states import ClientPublish



def test_ClientChannels() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    channels = ClientChannels()


    attrs = lattrs(channels)

    assert attrs == [
        '_ClientChannels__cached']


    assert inrepr(
        'states.ClientChannels',
        channels)

    assert isinstance(
        hash(channels), int)

    assert instr(
        'states.ClientChannels',
        channels)



def test_ClientChannels_basics() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    channels = ClientChannels()


    channels.create('test')

    select = (
        channels.select('test'))

    assert select is not None

    assert select.endumped == {
        'members': None,
        'title': None,
        'topic': None,
        'unique': 'test'}


    channels.set_title(
        'test', 'Test')

    channels.set_topic(
        'test', 'Test')

    channels.add_member(
        'test', 'mtmbot')

    select = (
        channels.select('Test'))

    assert select is not None

    assert select.endumped == {
        'members': {'mtmbot'},
        'title': 'Test',
        'topic': 'Test',
        'unique': 'test'}


    channels.delete('test')

    select = (
        channels.select('test'))

    assert select is None



def test_ClientChannels_cover() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    channels = ClientChannels()


    channels.create('test')

    channels.add_member(
        'test', 'mtmbot')

    channels.del_member(
        'test', 'mtmbot')

    select = (
        channels.select('test'))

    assert select is not None

    assert select.endumped == {
        'members': None,
        'title': None,
        'topic': None,
        'unique': 'test'}


    channels.add_member(
        'test', 'mtmbot')

    (channels
     .clear_members('test'))

    assert select is not None

    assert select.endumped == {
        'members': None,
        'title': None,
        'topic': None,
        'unique': 'test'}


    channels.set_title(
        'test1', 'Test')

    channels.set_topic(
        'test2', 'Test')

    channels.add_member(
        'test3', 'mtmbot')

    channels.rename_member(
        'mtmbot', 'test')


    select = (
        channels.select('test3'))

    assert select is not None

    assert select.endumped == {
        'members': {'test'},
        'title': None,
        'topic': None,
        'unique': 'test3'}



def test_ClientPublish() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    publish = ClientPublish()


    attrs = lattrs(publish)

    assert attrs == [
        '_ClientPublish__callbacks']


    assert inrepr(
        'states.ClientPublish',
        publish)

    assert isinstance(
        hash(publish), int)

    assert instr(
        'states.ClientPublish',
        publish)



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

from ..child import InvalidChild

if TYPE_CHECKING:
    from ...robie import Robie



def test_InvalidChild() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    raises = InvalidChild(
        child='invalid',
        phase='initial')


    attrs = lattrs(raises)

    assert attrs == [
        'child',
        'about']


    assert inrepr(
        'InvalidChild',
        raises)

    assert isinstance(
        hash(raises), int)

    assert instr(
        'Child (invalid)',
        raises)


    assert str(raises) == (
        'Child (invalid) '
        'invalid within '
        'phase (initial)')



def test_InvalidChild_cover(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    clients = childs.clients

    client = clients['ircbot']


    raises = InvalidChild(
        child=client,
        phase='runtime',
        about='about')

    name = client.name

    assert str(raises) == (
        f'Child ({name}) '
        'invalid within phase '
        '(runtime) (about)')



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from ...utils import ClientChannel
from ...utils import ClientChannels as _ClientChannels



class ClientChannels(_ClientChannels):
    """
    Cache the information regarding client status on server.
    """


    def create(
        self,
        unique: str,
    ) -> None:
        """
        Create the channel within the internal cache dictionary.

        :param unique: Unique identifier for channel on server.
        """

        cached = self.__cached

        item = ClientChannel(
            unique=unique,
            title=unique)

        cached[unique] = item



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .params import LoggerPluginParams
from .plugin import LoggerPlugin



__all__ = [
    'LoggerPlugin',
    'LoggerPluginParams']



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .child import InvalidChild
from .importer import importer
from .param import InvalidParam
from .states import ClientChannel
from .states import ClientChannels
from .states import ClientPublish
from .thread import DupliThread



__all__ = [
    'InvalidChild',
    'InvalidParam',
    'DupliThread',
    'ClientChannel',
    'ClientChannels',
    'ClientPublish',
    'importer']



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.types import inlist
from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

from ..client import RobieClientThread

if TYPE_CHECKING:
    from ...service import RobieService



def test_RobieClientThread(
    service: 'RobieService',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param service: Ancilary Chatting Robie class instance.
    """

    member = service.clients

    threads = (
        member.threads
        .values())


    for thread in threads:

        assert isinstance(
            thread,
            RobieClientThread)


        attrs = lattrs(thread)

        # Inherits Thread class

        assert inlist(
            '_RobieThread__member',
            attrs)

        assert inlist(
            '_RobieThread__child',
            attrs)

        assert inlist(
            '_RobieThread__mqueue',
            attrs)

        assert inlist(
            '_RobieThread__cqueue',
            attrs)


        assert inrepr(
            'Thread(RobieClient',
            thread)

        assert isinstance(
            hash(thread), int)

        assert instr(
            'Thread(RobieClient',
            thread)


        assert thread.robie

        assert thread.service

        assert thread.member

        assert thread.child

        assert thread.client

        assert thread.mqueue

        assert thread.cqueue

        assert not thread.congest

        assert not thread.enqueue



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .client import MTMClient
from .params import MTMClientParams



__all__ = [
    'MTMClient',
    'MTMClientParams']



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Any
from typing import Callable
from typing import Optional

from pydantic import Field

from .child import RobieChildParams
from .common import RobieParamsModel



class RobiePersonMatchParams(RobieParamsModel, extra='forbid'):
    """
    Process and validate the Robie configuration parameters.
    """

    client: Annotated[
        str,
        Field(...,
              description='Client where the user exists',
              min_length=1)]

    match: Annotated[
        list[str],
        Field(...,
              description='Values client uses to identify',
              min_length=1)]


    def __init__(
        # NOCVR
        self,
        /,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        match = data.get('match')

        if isinstance(match, str):
            data['match'] = [match]

        super().__init__(**data)



class RobiePersonParams(RobieChildParams, extra='forbid'):
    """
    Process and validate the Robie configuration parameters.
    """

    first: Annotated[
        Optional[str],
        Field(None,
              description='Additional optional information',
              min_length=1)]

    last: Annotated[
        Optional[str],
        Field(None,
              description='Additional optional information',
              min_length=1)]

    about: Annotated[
        Optional[str],
        Field(None,
              description='Additional optional information',
              min_length=1)]

    matches: Annotated[
        list[RobiePersonMatchParams],
        Field(...,
              description='How the user will be identified')]

    weight: Annotated[
        int,
        Field(50,
              description='Determine order of precedence',
              ge=1, le=100)]


    def __init__(
        # NOCVR
        self,
        /,
        _parse: Optional[Callable[..., Any]] = None,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """


        if _parse is not None:

            parsable = [
                'first',
                'last',
                'about',
                'matches',
                'weight']

            for key in parsable:

                value = data.get(key)

                if value is None:
                    continue

                value = _parse(value)

                data[key] = value


        super().__init__(**data)



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ...robie.models import RobieMessage
    from .plugin import HomiePlugin



def composedsc(
    plugin: 'HomiePlugin',
    mitem: 'RobieMessage',
) -> None:
    """
    Construct and format message for related chat platform.

    :param plugin: Plugin class instance for Chatting Robie.
    :param mitem: Item containing information for operation.
    :param status: Object containing the status information.
    """

    from ...clients.discord.message import DSCMessage


    assert plugin.thread

    thread = plugin.thread
    robie = plugin.robie
    member = thread.member
    cqueue = member.cqueue
    persist = plugin.persist

    assert isinstance(
        mitem, DSCMessage)

    assert mitem.message

    unique = (
        mitem.message
        .split(' ', 1)[1])

    current = (
        persist
        .record(unique))

    citem = mitem.reply(
        robie, str(current))

    cqueue.put(citem)



def composeirc(
    plugin: 'HomiePlugin',
    mitem: 'RobieMessage',
) -> None:
    """
    Construct and format message for related chat platform.

    :param plugin: Plugin class instance for Chatting Robie.
    :param mitem: Item containing information for operation.
    :param status: Object containing the status information.
    """

    from ...clients.irc.message import IRCMessage


    assert plugin.thread

    thread = plugin.thread
    robie = plugin.robie
    member = thread.member
    cqueue = member.cqueue
    persist = plugin.persist

    assert isinstance(
        mitem, IRCMessage)

    assert mitem.message

    unique = (
        mitem.message
        .split(' ', 1)[1])

    current = (
        persist
        .record(unique))

    citem = mitem.reply(
        robie, str(current))

    cqueue.put(citem)



def composemtm(
    plugin: 'HomiePlugin',
    mitem: 'RobieMessage',
) -> None:
    """
    Construct and format message for related chat platform.

    :param plugin: Plugin class instance for Chatting Robie.
    :param mitem: Item containing information for operation.
    :param status: Object containing the status information.
    """

    from ...clients.mattermost.message import MTMMessage


    assert plugin.thread

    thread = plugin.thread
    robie = plugin.robie
    member = thread.member
    cqueue = member.cqueue
    persist = plugin.persist

    assert isinstance(
        mitem, MTMMessage)

    assert mitem.message

    unique = (
        mitem.message
        .split(' ', 1)[1])

    current = (
        persist
        .record(unique))

    citem = mitem.reply(
        robie, str(current))

    cqueue.put(citem)



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .clients import RobieClients
from .member import RobieMember
from .plugins import RobiePlugins



__all__ = [
    'RobieMember',
    'RobieClients',
    'RobiePlugins']



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Any
from typing import Callable
from typing import Optional

from pydantic import Field

from ..status import StatusPluginIconParams
from ...robie.params import RobiePluginParams
from ...robie.params.common import RobieParamsModel



class HomiePluginCommandParams(RobieParamsModel, extra='forbid'):
    """
    Process and validate the Robie configuration parameters.
    """

    irc: Annotated[
        str,
        Field('!persist',
              description='Command name for chat platform',
              min_length=2)]

    dsc: Annotated[
        str,
        Field('!persist',
              description='Command name for chat platform',
              min_length=2)]

    mtm: Annotated[
        str,
        Field('!persist',
              description='Command name for chat platform',
              min_length=2)]



class HomiePluginParams(RobiePluginParams, extra='forbid'):
    """
    Process and validate the Robie configuration parameters.
    """

    command: Annotated[
        HomiePluginCommandParams,
        Field(default_factory=HomiePluginCommandParams,
              description='Command name per chat platform')]

    restful: Annotated[
        str,
        Field(...,
              description='Where to find the RESTful API',
              examples=['http://localhost:8420'],
              min_length=1)]

    username: Annotated[
        Optional[str],
        Field(None,
              description='Authenticate with the service',
              min_length=1)]

    password: Annotated[
        Optional[str],
        Field(None,
              description='Authenticate with the service',
              min_length=1)]

    timeout: Annotated[
        int,
        Field(30,
              description='Timeout connecting to server',
              ge=1, le=300)]

    ssl_verify: Annotated[
        bool,
        Field(True,
              description='Verify the ceritifcate valid')]

    ssl_capem: Annotated[
        Optional[str],
        Field(None,
              description='Verify the ceritifcate valid',
              min_length=1)]

    clients: Annotated[
        list[str],
        Field(...,
              description='List of clients to enable plugin',
              min_length=1)]

    trusted: Annotated[
        Optional[list[str]],
        Field(None,
              description='Users are trusted by the plugin',
              min_length=1)]

    status: Annotated[
        StatusPluginIconParams,
        Field(default_factory=StatusPluginIconParams,
              description='Icon used per chat platform')]


    def __init__(
        # NOCVR
        self,
        /,
        _parse: Optional[Callable[..., Any]] = None,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """


        if _parse is not None:

            parsable = [
                'restful',
                'username',
                'password',
                'timeout',
                'ssl_verify',
                'ssl_capem',
                'status']

            for key in parsable:

                value = data.get(key)

                if value is None:
                    continue

                value = _parse(value)

                data[key] = value


        super().__init__(**data)



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

if TYPE_CHECKING:
    from ...robie import Robie



def test_RobiePerson(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    persons = childs.persons
    clients = childs.clients


    person = persons['hubert']


    attrs = lattrs(person)

    assert attrs == [
        '_RobieChild__robie',
        '_RobieChild__name',
        '_RobieChild__params']


    assert inrepr(
        'person.RobiePerson',
        person)

    assert isinstance(
        hash(person), int)

    assert instr(
        'person.RobiePerson',
        person)


    person.validate()

    assert person.robie

    assert person.enable

    assert person.name == 'hubert'

    assert person.kind == 'person'

    assert person.params

    assert person.dumped

    assert person.first == 'Hubert'

    assert person.last == 'Farnsworth'

    assert person.about

    assert person.weight


    assert person.match(
        clients['dscbot'],
        '823039201390230492')

    assert not person.match(
        clients['dscbot'],
        '823902304920392013')


    assert not person.match(
        clients['ircbot'],
        'anonymous')

    assert person.match(
        clients['ircbot'],
        'hubert!hubert@science.com')

    assert not person.match(
        clients['ircbot'],
        'bender!bender@bending.com')


    assert person.match(
        clients['mtmbot'],
        'kjf9al2klaiietalkw')

    assert not person.match(
        clients['mtmbot'],
        'iietalkwkjf9al2kla')



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING
from typing import Type

from .child import RobieChild

if TYPE_CHECKING:
    from ..params import RobiePluginParams
    from ..threads import RobiePluginThread



class RobiePlugin(RobieChild):
    """
    Integrate with the Robie routine and perform operations.
    """

    __thread: Optional['RobiePluginThread'] = None


    @property
    def kind(
        self,
    ) -> Literal['plugin']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'plugin'


    @classmethod
    def schema(
        cls,
    ) -> Type['RobiePluginParams']:
        """
        Return the configuration parameters relevant for class.

        :returns: Configuration parameters relevant for class.
        """

        raise NotImplementedError


    @property
    def thread(
        self,
    ) -> Optional['RobiePluginThread']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        from ..threads import (
            RobiePluginThread)

        thread = self.__thread

        if thread is None:
            return None

        assert isinstance(
            thread,
            RobiePluginThread)

        return thread


    @thread.setter
    def thread(
        self,
        value: 'RobiePluginThread',
    ) -> None:
        """
        Update the value for the attribute from class instance.
        """

        from ..threads import (
            RobiePluginThread)

        assert isinstance(
            value,
            RobiePluginThread)

        self.__thread = value


    def operate(
        self,
    ) -> None:
        """
        Perform the operation related to Robie service threads.

        .. note::
           Deviates from enhomie in children have operations,
           and are more isolated from internal core routines.
        """

        raise NotImplementedError



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.times import Time

from ...childs import RobieClient
from ...models import RobieCommand
from ...models import RobieMessage

if TYPE_CHECKING:
    from ...service import RobieService



def test_RobieThread_cover(
    service: 'RobieService',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param service: Ancilary Chatting Robie class instance.
    """

    member = service.clients

    message = RobieMessage
    command = RobieCommand


    threads = (
        member.threads
        .values())

    for thread in threads:

        child = thread.child

        assert isinstance(
            child, RobieClient)


        mitem = message(child)

        expired = (
            thread.expired(mitem))

        assert expired is False

        mitem.time = Time('-1h')

        expired = (
            thread.expired(mitem))

        assert expired is True


        citem = command(child)

        expired = (
            thread.expired(citem))

        assert expired is False

        citem.time = Time('-1h')

        expired = (
            thread.expired(citem))

        assert expired is True



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from threading import Thread
from time import sleep as block_sleep
from typing import TYPE_CHECKING

from encommon.times import Timer
from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

from enconnect.fixtures import IRCClientSocket

from ..plugin import AutoJoinPlugin
from ....clients.irc.test import IRCEVENTS

if TYPE_CHECKING:
    from ....robie import RobieService



IRCEVENTS = deepcopy(IRCEVENTS)



def test_AutoJoinPlugin(
    service: 'RobieService',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param service: Ancilary Chatting Robie class instance.
    """

    plugin = (
        service.plugins
        .childs['autojoin'])

    assert isinstance(
        plugin, AutoJoinPlugin)


    attrs = lattrs(plugin)

    assert attrs == [
        '_RobieChild__robie',
        '_RobieChild__name',
        '_RobieChild__params',
        '_AutoJoinPlugin__started',
        '_AutoJoinPlugin__should',
        '_AutoJoinPlugin__joined',
        '_AutoJoinPlugin__timer',
        '_RobiePlugin__thread']


    assert inrepr(
        'plugin.AutoJoinPlugin',
        plugin)

    assert isinstance(
        hash(plugin), int)

    assert instr(
        'plugin.AutoJoinPlugin',
        plugin)


    plugin.validate()

    assert plugin.robie

    assert plugin.enable

    assert plugin.name == 'autojoin'

    assert plugin.kind == 'plugin'

    assert plugin.schema()

    assert plugin.params

    assert plugin.thread

    assert plugin.dumped



def test_AutoJoinPlugin_cover(
    service: 'RobieService',
    client_ircsock: IRCClientSocket,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param service: Ancilary Chatting Robie class instance.
    :param client_ircsock: Object to mock client connection.
    """

    plugin = (
        service.plugins
        .childs['autojoin'])

    setattr(
        plugin,
        '_AutoJoinPlugin__timer',
        Timer(0))


    client_ircsock([

        (':mocked 376 ircbot '
         ':End of /MOTD command.'),

        ':ircbot JOIN :#enrobie',

        (':foo!bar@baz KICK '
         '#enrobie ircbot :foo'),

        ':ircbot JOIN :#enrobie',

        ':ircbot PART :#enrobie'])


    service.limit(
        plugins=[
            'autojoin',
            'status'])

    service.start()


    thread = Thread(
        target=service.operate)

    thread.start()


    block_sleep(5)

    service.soft()

    while service.running:
        block_sleep(1)

    service.stop()

    thread.join()



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from json import dumps
from pathlib import Path
from threading import Thread
from time import sleep as block_sleep
from typing import TYPE_CHECKING

from encommon.types import DictStrAny
from encommon.types import expate
from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

from enconnect.fixtures import MTMClientSocket
from enconnect.mattermost import ClientEvent
from enconnect.mattermost.test import EVENTS

from httpx import Response

from respx import MockRouter

from ..client import MTMClient
from ..command import MTMCommand
from ....conftest import config_factory
from ....conftest import robie_factory
from ....conftest import service_factory
from ....robie.addons import RobieQueue

if TYPE_CHECKING:
    from ....robie import Robie
    from ....robie import RobieService
    from ....robie.models import RobieCommand  # noqa: F401
    from ....robie.models import RobieMessage  # noqa: F401



MTMEVENT_RANDOM_CHAN = expate({
    'event': 'posted',
    'seq': 3,
    'broadcast/channel_id': 'enrobie',
    'data/channel_type': 'P',
    'data/post': (
        '{"user_id":"userid",'
        '"channel_id":"enrobie",'
        '"message":"Hello mtmbot"}'),
    'data/sender_name': '@user'})

MTMEVENT_RANDOM_PRIV = expate({
    'event': 'posted',
    'seq': 4,
    'broadcast/channel_id': 'privid',
    'data/channel_type': 'D',
    'data/post': (
        '{"user_id":"userid",'
        '"channel_id":"privid",'
        '"message":"Hello mtmbot"}'),
    'data/sender_name': '@user'})

MTMEVENT_HUBERT_CHAN = expate({
    'event': 'posted',
    'seq': 6,
    'broadcast/channel_id': 'enrobie',
    'data/channel_type': 'P',
    'data/post': (
        '{"user_id":"kjf9al2klaiietalkw",'
        '"channel_id":"enrobie",'
        '"message":"mtmbot"}'),
    'data/sender_name': '@hubert'})

MTMEVENT_HUBERT_PRIV = expate({
    'event': 'posted',
    'seq': 7,
    'broadcast/channel_id': 'privid',
    'data/channel_type': 'D',
    'data/post': (
        '{"user_id":"kjf9al2klaiietalkw",'
        '"channel_id":"privid",'
        '"message":"mtmbot"}'),
    'data/sender_name': '@hubert'})

_MTMEVENTS: list[DictStrAny] = [

    {'event': 'hello',
     'broadcast/user_id': 'mtmunq'},

    {'event': 'channel_updated',
     'broadcast/channel_id': 'enrobie',
     'data/channel': (
         '{"id":"enrobie",'
         '"name":"enrobie",'
         '"header":"Test topic is changed"}'),
     'seq': 2},

    # From random to channel
    MTMEVENT_RANDOM_CHAN,

    # From random to private
    MTMEVENT_RANDOM_PRIV,

    # From hubert to channel
    MTMEVENT_HUBERT_CHAN,

    # From hubert to channel
    MTMEVENT_HUBERT_CHAN,

    # From hubert to private
    MTMEVENT_HUBERT_PRIV,

    {'event': 'channel_updated',
     'broadcast/channel_id': 'enrobie',
     'data/channel': (
         '{"id":"enrobie",'
         '"name":"enrobie",'
         '"header":"Test topic"}'),
     'seq': 8}]

MTMEVENTS = EVENTS + [
    expate(x) for x in _MTMEVENTS]



def test_MTMClient(
    service: 'RobieService',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param service: Ancilary Chatting Robie class instance.
    """

    client = (
        service.clients
        .childs['mtmbot'])

    assert isinstance(
        client,
        MTMClient)


    attrs = lattrs(client)

    assert attrs == [
        '_RobieChild__robie',
        '_RobieChild__name',
        '_RobieChild__params',
        '_MTMClient__client',
        '_MTMClient__channels',
        '_MTMClient__publish',
        '_RobieClient__thread']


    assert inrepr(
        'client.MTMClient',
        client)

    assert isinstance(
        hash(client), int)

    assert instr(
        'client.MTMClient',
        client)


    client.validate()

    assert client.robie

    assert client.enable

    assert client.name == 'mtmbot'

    assert client.family == 'mattermost'

    assert client.kind == 'client'

    assert client.client

    assert client.channels

    assert client.publish

    assert client.schema()

    assert client.params

    assert client.thread

    assert client.dumped



def test_MTMClient_message(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    clients = childs.clients

    client = clients['mtmbot']

    assert isinstance(
        client, MTMClient)


    _queue = RobieQueue[
        'RobieMessage']

    queue: _queue = (
        RobieQueue(robie))


    _event = {
        'status': 'OK',
        'seq_reply': 1}

    event = ClientEvent(
        client.client, _event)

    client.put_message(
        queue, event)



def test_MTMClient_command(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    clients = childs.clients

    client = clients['mtmbot']

    assert isinstance(
        client, MTMClient)


    _queue = RobieQueue[
        'RobieCommand']

    queue: _queue = (
        RobieQueue(robie))


    client.put_command(
        queue, 'delete',
        'posts/mocked')



def test_MTMClient_compose(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    clients = childs.clients

    client = clients['mtmbot']

    assert isinstance(
        client, MTMClient)


    citem = (
        client.compose(
            'chan', 'message'))

    assert isinstance(
        citem, MTMCommand)


    assert citem.method == 'post'

    assert citem.path == 'posts'

    assert not citem.params

    assert citem.json == {
        'channel_id': 'chan',
        'message': 'message'}



def test_MTMClient_channels(
    tmp_path: Path,
    client_mtmsock: MTMClientSocket,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param tmp_path: pytest object for temporal filesystem.
    :param client_mtmsock: Object to mock client connection.
    :param respx_mock: Object for mocking request operation.
    """

    robie = robie_factory(
        config_factory(tmp_path))

    service = (
        service_factory(robie))


    childs = robie.childs
    clients = childs.clients

    client = clients['mtmbot']

    assert isinstance(
        client, MTMClient)


    mocked = dumps([
        {'header': 'Testing',
         'id': 'enrobie',
         'name': 'enrobie'}])

    (respx_mock
     .get(
         'https://mocked:443'
         '/api/v4/users/mtmunq/'
         'teams/mocked/channels')
     .mock(Response(
         status_code=200,
         content=mocked)))


    client_mtmsock(MTMEVENTS)


    service.limit(
        clients=['mtmbot'],
        plugins=['status'])

    service.start()


    thread = Thread(
        target=service.operate)

    thread.start()


    block_sleep(5)


    select = (
        client.channels
        .select('enrobie'))

    assert select is not None

    assert select.endumped == {
        'members': None,
        'title': 'enrobie',
        'topic': 'Test topic',
        'unique': 'enrobie'}


    service.soft()

    while service.running:
        block_sleep(1)

    service.stop()

    thread.join()



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .jinja2 import RobieJinja2
from .logger import RobieLogger
from .queue import RobieQueue
from .queue import RobieQueueItem



__all__ = [
    'RobieLogger',
    'RobieJinja2',
    'RobieQueue',
    'RobieQueueItem']



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .params import AutoNickPluginParams
from .plugin import AutoNickPlugin



__all__ = [
    'AutoNickPlugin',
    'AutoNickPluginParams']



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from re import sub as re_sub
from typing import TYPE_CHECKING

from encommon.types import NCTrue
from encommon.utils import read_text
from encommon.utils import save_text
from encommon.utils.sample import ENPYRWS

from enconnect.irc import ClientEvent

from pydantic_ai.models.test import TestModel

from pytest import mark

from . import SAMPLES
from .test_history import _ainswer_history
from ..common import AinswerResponse
from ..history import AinswerHistoryKinds
from ..plugin import AinswerPlugin
from ...logger import LoggerPlugin
from ...logger.test.test_history import _logger_history
from ....clients.irc import IRCClient
from ....clients.irc.message import IRCMessage
from ....clients.irc.test import IRCEVENT_HUBERT_CHAN
from ....clients.irc.test import IRCEVENT_HUBERT_PRIV
from ....clients.irc.test import IRCEVENT_RANDOM_CHAN
from ....clients.irc.test import IRCEVENT_RANDOM_PRIV

if TYPE_CHECKING:
    from ....robie import Robie



def test_AinswerQuestion_engage(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    plugins = childs.plugins

    plugin = plugins['ainswer']

    assert isinstance(
        plugin, AinswerPlugin)

    question = plugin.question


    testing = TestModel()

    override_agent = (
        plugin.agent
        .override(
            model=testing))

    with override_agent:

        response = (
            question.submit(
                'Hello World!',
                AinswerResponse))

    assert response.text == 'a'



@mark.parametrize(
    'file,kind,event',
    [('random_chanmsg', 'chanmsg',
      IRCEVENT_RANDOM_CHAN),
     ('random_privmsg', 'privmsg',
      IRCEVENT_RANDOM_PRIV),
     ('hubert_chanmsg', 'chanmsg',
      IRCEVENT_HUBERT_CHAN),
     ('hubert_privmsg', 'privmsg',
      IRCEVENT_HUBERT_PRIV)])
def test_AinswerQuestion_prompt(
    robie: 'Robie',
    file: str,
    kind: 'AinswerHistoryKinds',
    event: str,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param file: Name of the file within the samples folder.
    :param kind: What kind of Robie message we dealing with.
    :param event: Raw event received from the network peer.
    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    clients = childs.clients
    plugins = childs.plugins

    client = clients['ircbot']
    plugin = plugins['ainswer']
    logger = plugins['logger']

    assert isinstance(
        client, IRCClient)

    assert isinstance(
        plugin, AinswerPlugin)

    assert isinstance(
        logger, LoggerPlugin)

    question = plugin.question


    item = IRCMessage(
        client,
        ClientEvent(
            client.client,
            event))

    assert item.kind == kind
    assert item.author
    assert item.author == (
        ('hubert', 'hubert')
        if 'hubert' in file
        else ('nick', 'nick'))

    if 'hubert' not in file:
        item.event.author = 'nickname1'
        item.event.recipient = (
            '#enrobie'
            if kind == 'chanmsg'
            else 'ircbot')

    item.event.whome = 'ircbot'

    item.event.message = (
        'Do the thing?')


    _ainswer_history(
        plugin, client)

    _logger_history(
        logger, client)


    sample_path = (
        SAMPLES / f'{file}.txt')

    prompt = (
        question.prompt(
            item,
            'Do the thing!'))

    prompt = re_sub(
        (r'\d{4}\-\d{2}\-\d{2}'
         r'T\d{2}:\d{2}:\d{2}'
         r'(\.\d+)?\+0000'),
        '1980-01-01T00:00:00Z',
        prompt)

    if ENPYRWS is NCTrue:
        save_text(
            sample_path,
            prompt)

    loaded = read_text(
        sample_path)

    assert prompt == loaded



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from argparse import ArgumentParser
from signal import SIGHUP
from signal import SIGINT
from signal import SIGTERM
from signal import signal
from sys import argv as sys_argv
from typing import Optional

from encommon.types import DictStrAny

from ..robie import Robie
from ..robie import RobieConfig
from ..robie import RobieService



def arguments(
    args: Optional[list[str]] = None,
) -> DictStrAny:
    """
    Construct arguments which are associated with the file.

    :param args: Override the source for the main arguments.
    :returns: Construct arguments from command line options.
    """

    parser = ArgumentParser()

    args = args or sys_argv[1:]


    parser.add_argument(
        '--config',
        help=(
            'complete or relative '
            'path to config file'))


    parser.add_argument(
        '--console',
        action='store_true',
        default=False,
        help=(
            'write log messages '
            'to standard output'))


    parser.add_argument(
        '--debug',
        action='store_true',
        default=False,
        help=(
            'increase logging level '
            'for standard output'))


    parser.add_argument(
        '--print_message',
        action='store_true',
        default=False,
        dest='pmessage',
        help=(
            'print the received '
            'messages to console'))


    parser.add_argument(
        '--print_command',
        action='store_true',
        default=False,
        dest='pcommand',
        help=(
            'print the submited '
            'commands to console'))


    return vars(
        parser
        .parse_args(args))



def operation(
    # NOCVR
    robie: Robie,
) -> None:
    """
    Perform whatever operation is associated with the file.

    :param robie: Primary class instance for Chatting Robie.
    """

    service = RobieService(robie)

    service.start()

    signal(SIGINT, service.soft)
    signal(SIGTERM, service.soft)
    signal(SIGHUP, service.soft)

    service.operate()



def execution(
    # NOCVR
    args: Optional[list[str]] = None,
) -> None:
    """
    Perform whatever operation is associated with the file.

    :param args: Override the source for the main arguments.
    """

    config = RobieConfig(
        arguments(args))

    config.logger.start()

    config.logger.log_i(
        base='execution/service',
        status='started')

    robie = Robie(config)

    operation(robie)

    config.logger.log_i(
        base='execution/service',
        status='stopped')

    config.logger.stop()



if __name__ == '__main__':
    execution()  # NOCVR



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from json import loads
from threading import Event
from threading import Thread
from threading import enumerate as thread_enumerate
from time import sleep as block_sleep
from typing import Any
from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING
from typing import Type

from encommon.types import NCNone

from enconnect.mattermost import Client
from enconnect.mattermost import ClientEvent
from enconnect.utils.http import _METHODS
from enconnect.utils.http import _PAYLOAD

from .command import MTMCommand
from .message import MTMMessage
from .params import MTMClientParams
from ...plugins import StatusPlugin
from ...plugins import StatusPluginStates
from ...robie.addons import RobieQueue
from ...robie.childs import RobieClient
from ...utils import ClientChannels
from ...utils import ClientPublish
from ...utils import DupliThread

if TYPE_CHECKING:
    from ...robie.models import RobieCommand
    from ...robie.models import RobieMessage



class MTMClient(RobieClient):
    """
    Establish and maintain connection with the chat service.
    """

    __client: Client
    __channels: ClientChannels
    __publish: ClientPublish


    def __post__(
        self,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        params = self.params

        client = Client(
            params.client,
            self.__debugger)

        self.__client = client

        self.__channels = (
            ClientChannels())

        self.__publish = (
            ClientPublish())


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        # Nothing to do for client


    @classmethod
    def schema(
        cls,
    ) -> Type[MTMClientParams]:
        """
        Return the configuration parameters relevant for class.

        :returns: Configuration parameters relevant for class.
        """

        return MTMClientParams


    @property
    def params(
        self,
    ) -> MTMClientParams:
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        params = super().params

        assert isinstance(
            params,
            MTMClientParams)

        return params


    @property
    def family(
        self,
    ) -> Literal['mattermost']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'mattermost'


    @property
    def channels(
        self,
    ) -> ClientChannels:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__channels


    @property
    def publish(
        self,
    ) -> ClientPublish:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__publish


    @property
    def client(
        self,
    ) -> Client:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__client


    def operate(  # noqa: CFQ001
        self,
    ) -> None:
        """
        Perform the operation related to Robie service threads.
        """

        assert self.thread

        thread = self.thread
        robie = thread.robie
        member = thread.member
        target = member.mqueue
        cqueue = thread.cqueue
        vacate = member.vacate
        params = self.params

        delay = params.delay

        self.__status('pending')


        def _put_mqueue() -> None:

            event = source.get()

            try:
                self.__event(event)

            except Exception as reason:

                robie.logger.log_e(
                    base=self,
                    name=self,
                    status='exception',
                    exc_info=reason)

            self.put_message(
                target, event)


        def _get_cqueue() -> None:

            citem = cqueue.get()

            assert isinstance(
                citem, MTMCommand)

            try:
                client.request(
                    method=citem.method,
                    path=citem.path,
                    params=citem.params,
                    json=citem.json,
                    timeout=10)

            except Exception as reason:

                robie.logger.log_e(
                    base=self,
                    name=self,
                    status='exception',
                    exc_info=reason)


        def _continue() -> bool:

            return (
                not vacate.is_set()
                and daerht.is_alive())


        client = self.__client
        source = client.mqueue


        def _operate() -> None:

            try:

                client.operate()

                self.__status('pending')

            except Exception as reason:

                robie.logger.log_e(
                    base=self,
                    name=self,
                    status='exception',
                    exc_info=reason)

                self.__status('failure')

            finally:
                pending.set()


        def _routine() -> None:

            while _continue():

                robie.logger.log_i(
                    base=self,
                    name=self,
                    status='connect')

                pending.set()

                _operate()

                robie.logger.log_i(
                    base=self,
                    name=self,
                    status='severed')

                block_sleep(delay)


        pending = Event()


        name = (
            f'{self.name}'
            '_thread_client')

        threads = (
            x.name for x in
            thread_enumerate())

        if name in threads:
            raise DupliThread(name)

        daerht = Thread(
            name=name,
            target=_routine)

        daerht.start()


        def _connected() -> bool:

            return all([
                client.connected,
                client.nickname])


        while _continue():

            if (pending.is_set()
                    and _connected()):

                self.__status('normal')

                pending.clear()

            if not cqueue.empty:
                _get_cqueue()

            block_sleep(0.025)

            while not source.empty():
                _put_mqueue()

            block_sleep(0.025)


        client.stop()

        while daerht.is_alive():
            daerht.join(1)

        daerht.join()

        while not source.empty():
            _put_mqueue()  # NOCVR


    def __event(
        self,
        event: ClientEvent,
    ) -> None:
        """
        Process the provided message item from the Robie thread.

        :param event: Raw event received from the network peer.
        """

        client = self.__client

        type = event.type
        data = event.data
        broadcast = event.broadcast


        if type == 'channel_updated':

            if data is None:
                return NCNone

            chan = loads(data['channel'])

            chid = chan['id']

            if 'header' in chan:

                topic = chan['header']

                (self.channels
                 .set_topic(chid, topic))

            if 'name' in chan:

                name = chan['name']

                (self.channels
                 .set_title(chid, name))


        if type == 'hello':

            assert broadcast

            uniq = broadcast['user_id']
            teamid = client.params.teamid

            path = (
                f'users/{uniq}/teams'
                f'/{teamid}/channels')

            chans = (
                client
                .request('get', path)
                .json())

            for chan in chans:

                chid = chan['id']

                if 'header' in chan:

                    topic = chan['header']

                    (self.channels
                     .set_topic(chid, topic))

                if 'name' in chan:

                    name = chan['name']

                    (self.channels
                     .set_title(chid, name))


    def get_message(
        self,
        event: Optional[ClientEvent] = None,
    ) -> MTMMessage:
        """
        Return the new item containing information for operation.

        .. note::
           Though parameters are marked optional, that is not
           necessarily the case; for typing purposes one cannot
           introduce required parameters outside of the parent.

        :param event: Raw event received from the network peer.
        :returns: New item containing information for operation.
        """

        assert event is not None

        return MTMMessage(self, event)


    def put_message(
        self,
        mqueue: RobieQueue['RobieMessage'],
        event: Optional[ClientEvent] = None,
    ) -> None:
        """
        Insert the new item containing information for operation.

        .. note::
           Though parameters are marked optional, that is not
           necessarily the case; for typing purposes one cannot
           introduce required parameters outside of the parent.

        :param mqueue: Queue instance where the item is received.
        :param event: Raw event received from the network peer.
        """

        mitem = self.get_message(event)

        mqueue.put(mitem)

        self.__publish.publish(mitem)


    def get_command(
        self,
        method: Optional[_METHODS] = None,
        path: Optional[str] = None,
        params: Optional[_PAYLOAD] = None,
        json: Optional[_PAYLOAD] = None,
    ) -> MTMCommand:
        """
        Return the new item containing information for operation.

        .. note::
           Though parameters are marked optional, that is not
           necessarily the case; for typing purposes one cannot
           introduce required parameters outside of the parent.

        :returns: New item containing information for operation.
        :param method: Method for operation with the API server.
        :param path: Path for the location to upstream endpoint.
        :param params: Optional parameters included in request.
        :param json: Optional JSON payload included in request.
        """

        assert method is not None
        assert path is not None

        return MTMCommand(
            self,
            method, path,
            params, json)


    def put_command(
        self,
        cqueue: RobieQueue['RobieCommand'],
        method: Optional[_METHODS] = None,
        path: Optional[str] = None,
        params: Optional[_PAYLOAD] = None,
        json: Optional[_PAYLOAD] = None,
    ) -> None:
        """
        Insert the new item containing information for operation.

        .. note::
           Though parameters are marked optional, that is not
           necessarily the case; for typing purposes one cannot
           introduce required parameters outside of the parent.

        :param cqueue: Queue instance where the item is received.
        :param method: Method for operation with the API server.
        :param path: Path for the location to upstream endpoint.
        :param params: Optional parameters included in request.
        :param json: Optional JSON payload included in request.
        """

        citem = (
            self.get_command(
                method, path,
                params, json))

        cqueue.put(citem)


    def compose(
        self,
        target: str,
        content: str,
    ) -> 'RobieCommand':
        """
        Compose the message and transmit using the chat client.

        .. note::
           Though parameters are marked optional, that is not
           necessarily the case; for typing purposes one cannot
           introduce required parameters outside of the parent.

        :param target: Where the composed message will be sent.
        :param content: Content that will be source of message.
        """

        assert target is not None

        payload = {
            'channel_id': target,
            'message': content}

        return (
            self.get_command(
                'post', 'posts',
                json=payload))


    def __debugger(
        self,
        **kwargs: Any,
    ) -> None:
        """
        Pass the provided keyword arguments into logger object.

        :param kwargs: Keyword arguments for populating message.
        """

        robie = self.robie

        assert not any([
            kwargs.get('level'),
            kwargs.get('base'),
            kwargs.get('name')])

        robie.logger.log_d(
            base=self,
            name=self,
            **kwargs)


    def __status(
        self,
        status: StatusPluginStates,
    ) -> None:
        """
        Update or insert the status of the Robie child instance.

        :param status: One of several possible value for status.
        """

        thread = self.thread
        params = self.params

        if thread is None:
            return NCNone

        plugins = (
            thread.service
            .plugins.childs)

        if 'status' not in plugins:
            return NCNone

        plugin = plugins['status']

        assert isinstance(
            plugin, StatusPlugin)

        (plugin.update(
            unique=self.name,
            group='Connections',
            title='Mattermost',
            icon=params.status,
            state=status))



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Optional
from typing import Type
from typing import Union

from encommon.types import NCNone

from .helpers import composedsc
from .helpers import composeirc
from .helpers import composemtm
from .params import HomiePluginParams
from .persist import HomiePersist
from ..ainswer import AinswerTool
from ..status import StatusPlugin
from ..status import StatusPluginStates
from ...robie.childs import RobiePerson
from ...robie.childs import RobiePlugin
from ...robie.models import RobieMessage



class HomiePlugin(RobiePlugin):
    """
    Integrate with the Robie routine and perform operations.

    .. note::
       This plugin allows for connecting to Homie Automate.
    """

    __started: bool

    __persist: HomiePersist


    def __post__(
        self,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__started = False

        self.__persist = (
            HomiePersist(self))

        self.__status('pending')


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        # Review the parameters


    @classmethod
    def schema(
        cls,
    ) -> Type[HomiePluginParams]:
        """
        Return the configuration parameters relevant for class.

        :returns: Configuration parameters relevant for class.
        """

        return HomiePluginParams


    @property
    def params(
        self,
    ) -> HomiePluginParams:
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        params = super().params

        assert isinstance(
            params,
            HomiePluginParams)

        return params


    @property
    def persist(
        self,
    ) -> HomiePersist:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__persist


    def operate(
        self,
    ) -> None:
        """
        Perform the operation related to Robie service threads.

        :param thread: Child class instance for Chatting Robie.
        """

        assert self.thread

        thread = self.thread
        mqueue = thread.mqueue
        params = self.params

        command = params.command

        names = (
            self.params.clients)


        match: Optional[str]

        kinds = ['privmsg', 'chanmsg']


        if not self.__started:
            self.__started = True
            self.__status('normal')


        while not mqueue.empty:

            mitem = mqueue.get()

            name = mitem.client
            time = mitem.time
            kind = mitem.kind
            family = mitem.family
            isme = mitem.isme
            message = mitem.message

            # Ignore unrelated events
            if kind not in kinds:
                continue

            # Ignore unrelated clients
            if name not in names:
                continue  # NOCVR

            # Ignore event from client
            if isme is True:
                continue

            # Ignore expired messages
            if time.since > 15:
                continue  # NOCVR

            # Basic trust enforcement
            if self.notrust(mitem):
                continue

            assert message is not None


            match = None

            if family == 'discord':
                match = command.dsc

            if family == 'irc':
                match = command.irc

            if family == 'mattermost':
                match = command.mtm

            # Ignore unrelated clients
            if match is None:
                continue  # NOCVR

            # Ignore unrelated message
            if not message.startswith(match):
                continue


            if family == 'discord':
                composedsc(self, mitem)

            if family == 'irc':
                composeirc(self, mitem)

            if family == 'mattermost':
                composemtm(self, mitem)


    def trusted(
        self,
        check: Union[str, RobiePerson, RobieMessage],
    ) -> bool:
        """
        Return the boolean indicating whether person is trusted.

        :param check: Validate the person is trusted by plugin.
        :returns: Boolean indicating whether person is trusted.
        """

        params = self.params
        trusted = params.trusted

        if trusted is None:
            return True

        if isinstance(check, RobieMessage):

            if not check.person:
                return False

            check = check.person

        elif isinstance(check, RobiePerson):
            check = check.name

        return check in trusted


    def notrust(
        self,
        check: Union[str, RobiePerson, RobieMessage],
    ) -> bool:
        """
        Return the boolean indicating whether person is trusted.

        :param check: Validate the person is trusted by plugin.
        :returns: Boolean indicating whether person is trusted.
        """

        return not self.trusted(check)


    def __status(
        self,
        status: StatusPluginStates,
    ) -> None:
        """
        Update or insert the status of the Robie child instance.

        :param status: One of several possible value for status.
        """

        thread = self.thread
        params = self.params

        if thread is None:
            return None

        plugins = (
            thread.service
            .plugins.childs)

        if 'status' not in plugins:
            return NCNone

        plugin = plugins['status']

        assert isinstance(
            plugin, StatusPlugin)

        (plugin.update(
            unique=self.name,
            group='Extensions',
            title='Homie Automate',
            icon=params.status,
            state=status))


    def ainswer(
        self,
    ) -> list[AinswerTool]:
        """
        Return the Ainswer tools that are related to the plugin.

        :returns: Ainswer tools that are related to the plugin.
        """

        from .ainswer import homie_persist

        return [homie_persist]



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from random import randint
from time import sleep as block_sleep
from typing import Optional
from typing import TYPE_CHECKING
from typing import Type
from typing import Union

from encommon.types import DictStrAny
from encommon.types import NCNone

from .ainswer import AinswerDepends
from .ainswer import AinswerQuestion
from .ainswer import AinswerToolset
from .common import AinswerResponse
from .helpers import composedsc
from .helpers import composeirc
from .helpers import composemtm
from .history import AinswerHistory
from .params import AinswerPluginParams
from ..status import StatusPlugin
from ..status import StatusPluginStates
from ...robie.childs import RobiePerson
from ...robie.childs import RobiePlugin
from ...robie.models import RobieMessage

if TYPE_CHECKING:
    from pydantic_ai import Agent
    from pydantic_ai.models import Model



class AinswerPlugin(RobiePlugin):
    """
    Integrate with the Robie routine and perform operations.

    .. note::
       This plugin allows for interacting with an LLM model.
    """

    __started: bool

    __toolset: AinswerToolset
    __question: AinswerQuestion
    __history: AinswerHistory

    __model: 'Model'
    __agent: Optional['Agent[AinswerDepends, str]']


    def __post__(
        self,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__started = False

        params = self.params

        ainswer = params.ainswer
        secret = ainswer.secret
        origin = ainswer.origin


        self.__toolset = (
            AinswerToolset(self))

        self.__question = (
            AinswerQuestion(self))

        self.__history = (
            AinswerHistory(self))


        model: 'Model' | None = None

        if origin == 'anthropic':

            from pydantic_ai.models import anthropic

            _anthropic = (
                anthropic.AnthropicModel)

            model = _anthropic(
                ainswer.model,
                api_key=secret)

        elif origin == 'openai':

            from pydantic_ai.models import openai

            _openai = (
                openai.OpenAIModel)

            model = _openai(
                ainswer.model,
                api_key=secret)

        assert model is not None, (
            'Model not instantiated')

        self.__model = model

        self.__agent = None


        self.__status('pending')


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        # Review the parameters


    @classmethod
    def schema(
        cls,
    ) -> Type[AinswerPluginParams]:
        """
        Return the configuration parameters relevant for class.

        :returns: Configuration parameters relevant for class.
        """

        return AinswerPluginParams


    @property
    def params(
        self,
    ) -> AinswerPluginParams:
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        params = super().params

        assert isinstance(
            params,
            AinswerPluginParams)

        return params


    @property
    def toolset(
        self,
    ) -> AinswerToolset:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__toolset


    @property
    def question(
        self,
    ) -> AinswerQuestion:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__question


    @property
    def history(
        self,
    ) -> AinswerHistory:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__history


    @property
    def agent(
        self,
    ) -> 'Agent[AinswerDepends, str]':
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        from pydantic_ai import Agent
        from pydantic_ai.settings import ModelSettings

        agent = self.__agent

        if agent is not None:
            return agent

        params = self.params
        prompt = params.prompt
        ainswer = params.ainswer
        system = prompt.system

        toolset = (
            self.__toolset.toolset)

        settings = ModelSettings(
            timeout=ainswer.timeout)

        self.__agent = Agent(
            self.__model,
            system_prompt=system,
            model_settings=settings,
            deps_type=AinswerDepends,
            tools=toolset)

        return self.__agent


    def operate(
        self,
    ) -> None:
        """
        Perform the operation related to Robie service threads.
        """

        assert self.thread

        thread = self.thread
        mqueue = thread.mqueue

        names = (
            self.params.clients)


        if not self.__started:
            self.__started = True
            self.__status('normal')


        while not mqueue.empty:

            mitem = mqueue.get()

            name = mitem.client
            time = mitem.time
            isme = mitem.isme
            family = mitem.family

            # Ignore unrelated clients
            if name not in names:
                continue  # NOCVR

            # Ignore event from client
            if isme is True:
                continue

            # Ignore expired messages
            if time.since > 15:
                continue  # NOCVR

            # Basic trust enforcement
            if self.notrust(mitem):
                continue


            if family == 'discord':
                composedsc(self, mitem)

            if family == 'irc':
                composeirc(self, mitem)

            if family == 'mattermost':
                composemtm(self, mitem)


    def ainswer(  # noqa: CFQ001
        self,
        mitem: 'RobieMessage',
        prompt: str,
        respond: Type[AinswerResponse],
    ) -> str:
        """
        Submit the question to the LLM and return the response.

        :param mitem: Item containing information for operation.
        :param prompt: Additional prompt insert before question.
        :param respond: Model to describe the expected response.
        :returns: Response adhering to provided specifications.
        """

        robie = self.robie
        config = robie.config
        childs = robie.childs
        persons = childs.persons
        clients = childs.clients
        sargs = config.sargs
        question = self.question
        history = self.history
        params = self.params

        _ainswer = params.ainswer
        _prompt = params.prompt

        sleep = _ainswer.sleep
        system = _prompt.system

        assert mitem.whome
        assert mitem.author
        assert mitem.anchor
        assert mitem.message

        _client = mitem.client
        _person = mitem.person
        author = mitem.author
        anchor = mitem.anchor
        message = mitem.message

        client = (
            clients[_client]
            if _client is not None
            else None)

        person = (
            persons[_person]
            if _person is not None
            else None)


        robie.logger.log_i(
            base=self,
            name=self,
            item='ainswer',
            client=client.name,
            person=(
                person.name
                if person is not None
                else None),
            author=author[0],
            anchor=anchor,
            message=message)


        imsorry = (
            f"I'm sorry {author[0]}, I'm"
            " afraid I can't do that.")


        _sleep = randint(*sleep)

        robie.logger.log_i(
            base=self,
            name=self,
            item='ainswer',
            client=client.name,
            person=(
                person.name
                if person is not None
                else None),
            author=author[0],
            anchor=anchor,
            status='pausing',
            seconds=_sleep)

        # Useful to prevent abuse but
        # also reduce immediate reply
        # and allow for plugin events
        # before Jinja2 referencing
        block_sleep(_sleep)


        try:

            prompt = (
                question.prompt(
                    mitem, prompt))

            if sargs.get('console'):
                robie.printer({
                    'system': system,
                    'prompt': prompt,
                    'sleep': _sleep})

            response = (
                question.submit(
                    prompt, respond,
                    mitem=mitem))

            ainswer = response.text

            self.__status('normal')

            robie.logger.log_i(
                base=self,
                name=self,
                item='ainswer',
                client=client.name,
                person=(
                    person.name
                    if person is not None
                    else None),
                author=author[0],
                anchor=anchor,
                ainswer=ainswer)

            history.process(
                mitem, ainswer)

            return ainswer


        except Exception as reason:

            self.__status('failure')

            robie.logger.log_e(
                base=self,
                name=self,
                item='ainswer',
                client=client.name,
                person=(
                    person.name
                    if person is not None
                    else None),
                author=author[0],
                anchor=anchor,
                status='exception',
                exc_info=reason)

            return imsorry


    def printer(
        self,
        source: DictStrAny,
        color: int = 6,
    ) -> None:
        """
        Print the contents for the object within Robie instance.

        :param source: Content which will be shown after header.
        :param color: Override the color used for box character.
        """

        robie = self.robie
        config = robie.config
        sargs = config.sargs

        if not sargs.get('console'):
            return None

        robie.printer(source, color)


    def trusted(
        self,
        check: Union[str, RobiePerson, RobieMessage],
    ) -> bool:
        """
        Return the boolean indicating whether person is trusted.

        :param check: Validate the person is trusted by plugin.
        :returns: Boolean indicating whether person is trusted.
        """

        params = self.params
        trusted = params.trusted

        if trusted is None:
            return True

        if isinstance(check, RobieMessage):

            if not check.person:
                return False

            check = check.person

        elif isinstance(check, RobiePerson):
            check = check.name

        return check in trusted


    def notrust(
        self,
        check: Union[str, RobiePerson, RobieMessage],
    ) -> bool:
        """
        Return the boolean indicating whether person is trusted.

        :param check: Validate the person is trusted by plugin.
        :returns: Boolean indicating whether person is trusted.
        """

        return not self.trusted(check)


    def __status(
        self,
        status: StatusPluginStates,
    ) -> None:
        """
        Update or insert the status of the Robie child instance.

        :param status: One of several possible value for status.
        """

        thread = self.thread
        params = self.params

        if thread is None:
            return None

        plugins = (
            thread.service
            .plugins.childs)

        if 'status' not in plugins:
            return NCNone

        plugin = plugins['status']

        assert isinstance(
            plugin, StatusPlugin)

        (plugin.update(
            unique=self.name,
            group='Extensions',
            title='Ainswer',
            icon=params.status,
            state=status))



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path

from encommon.types import DictStrAny
from encommon.utils import save_text

from enconnect.fixtures import client_dscsock
from enconnect.fixtures import client_ircsock
from enconnect.fixtures import client_mtmsock

from pytest import fixture

from . import EXAMPLES
from . import PROJECT
from .robie import Robie
from .robie import RobieConfig
from .robie import RobieService



__all__ = [
    'client_dscsock',
    'client_ircsock',
    'client_mtmsock']



def config_factory(
    tmp_path: Path,
) -> RobieConfig:
    """
    Construct the instance for use in the downstream tests.

    :param tmp_path: pytest object for temporal filesystem.
    :returns: Newly constructed instance of related class.
    """

    content = (
        f"""

        enconfig:
          paths:
            - {EXAMPLES}
            - {tmp_path}/robie

        enlogger:
          stdo_level: info

        database: >-
          sqlite:///{tmp_path}/db

        """)

    config_path = (
        tmp_path / 'config.yml')

    save_text(
        config_path, content)

    sargs = {
        'config': config_path,
        'console': True,
        'debug': True,
        'pmessage': True,
        'pcommand': True}

    return RobieConfig(sargs)



@fixture
def config(
    tmp_path: Path,
) -> RobieConfig:
    """
    Construct the instance for use in the downstream tests.

    :param tmp_path: pytest object for temporal filesystem.
    :returns: Newly constructed instance of related class.
    """

    return config_factory(tmp_path)



@fixture
def replaces(
    tmp_path: Path,
) -> DictStrAny:
    """
    Return the complete mapping of what replaced in sample.

    :param tmp_path: pytest object for temporal filesystem.
    :returns: Complete mapping of what replaced in sample.
    """

    return {
        'PROJECT': PROJECT,
        'TMPPATH': tmp_path}



def robie_factory(
    config: RobieConfig,
) -> Robie:
    """
    Construct the instance for use in the downstream tests.

    :param config: Primary class instance for configuration.
    :returns: Newly constructed instance of related class.
    """

    return Robie(config)



@fixture
def robie(
    config: RobieConfig,
) -> Robie:
    """
    Construct the instance for use in the downstream tests.

    :param config: Primary class instance for configuration.
    :returns: Newly constructed instance of related class.
    """

    return robie_factory(config)



def service_factory(
    robie: Robie,
) -> RobieService:
    """
    Construct the instance for use in the downstream tests.

    :param robie: Primary class instance for Chatting Robie.
    :returns: Newly constructed instance of related class.
    """

    return RobieService(robie)



@fixture
def service(
    robie: Robie,
) -> RobieService:
    """
    Construct the instance for use in the downstream tests.

    :param robie: Primary class instance for Chatting Robie.
    :returns: Newly constructed instance of related class.
    """

    return service_factory(robie)



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING

from encommon.utils import fuzz_match

from .child import RobieChild
from ..models import RobieModels

if TYPE_CHECKING:
    from .client import RobieClient
    from ..params import RobiePersonParams



class RobiePerson(RobieChild):
    """
    Contain the properties regarding the actual user person.
    """


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        # Review the parameters


    @property
    def kind(
        self,
    ) -> Literal['person']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'person'


    @property
    def params(
        self,
    ) -> 'RobiePersonParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        model = (
            RobieModels
            .person())

        params = super().params

        assert isinstance(
            params, model)

        return params


    @property
    def first(
        self,
    ) -> Optional[str]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.params.first


    @property
    def last(
        self,
    ) -> Optional[str]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.params.last


    @property
    def about(
        self,
    ) -> Optional[str]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.params.about


    @property
    def weight(
        self,
    ) -> int:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.params.weight


    def match(
        self,
        client: 'RobieClient',
        check: str,
    ) -> bool:
        """
        Return the boolean indicating whether check is a match.

        :param client: Client class instance for Chatting Robie.
        :param check: Value to be searched within the haystack.
        :returns: Boolean indicating whether check is a match.
        """

        params = self.params

        name = client.name
        matches = params.matches

        matched: set[bool] = set()

        for match in matches:

            name = match.client
            patterns = match.match

            if client.name != name:
                continue

            fuzz = fuzz_match(
                check, patterns)

            matched.add(fuzz)

        return any(matched)



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .ainswer import AinswerPlugin
from .ainswer import AinswerPluginParams
from .autojoin import AutoJoinPlugin
from .autojoin import AutoJoinPluginParams
from .autonick import AutoNickPlugin
from .autonick import AutoNickPluginParams
from .enhomie import HomiePlugin
from .enhomie import HomiePluginParams
from .logger import LoggerPlugin
from .logger import LoggerPluginParams
from .nagios import NagiosPlugin
from .nagios import NagiosPluginParams
from .status import StatusPlugin
from .status import StatusPluginIconParams
from .status import StatusPluginItem
from .status import StatusPluginParams
from .status import StatusPluginStates



__all__ = [
    'AutoJoinPlugin',
    'AutoJoinPluginParams',
    'AutoNickPlugin',
    'AutoNickPluginParams',
    'AinswerPlugin',
    'AinswerPluginParams',
    'HomiePlugin',
    'HomiePluginParams',
    'NagiosPlugin',
    'NagiosPluginParams',
    'LoggerPlugin',
    'LoggerPluginParams',
    'StatusPlugin',
    'StatusPluginParams',
    'StatusPluginIconParams',
    'StatusPluginItem',
    'StatusPluginStates']



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from .thread import RobieThread

if TYPE_CHECKING:
    from ..childs import RobieClient
    from ..members import RobieClients



class RobieClientThread(RobieThread):
    """
    Common methods and routines for Chatting Robie threads.
    """


    @property
    def member(
        self,
    ) -> 'RobieClients':
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        from ..members import (
            RobieClients)

        member = super().member

        assert isinstance(
            member, RobieClients)

        return member


    @property
    def child(
        self,
    ) -> 'RobieClient':
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        from ..childs import (
            RobieClient)

        child = super().child

        assert isinstance(
            child, RobieClient)

        return child


    @property
    def client(
        self,
    ) -> 'RobieClient':
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.child



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Any
from typing import TYPE_CHECKING

from encommon.types import BaseModel

from enconnect.utils import HTTPClient

from enhomie.homie.addons.persist import _FIELD_ABOUT
from enhomie.homie.addons.persist import _FIELD_ABOUT_LABEL
from enhomie.homie.addons.persist import _FIELD_EXPIRE
from enhomie.homie.addons.persist import _FIELD_LEVEL
from enhomie.homie.addons.persist import _FIELD_TAGS
from enhomie.homie.addons.persist import _FIELD_UNIQUE
from enhomie.homie.addons.persist import _FIELD_UPDATE
from enhomie.homie.addons.persist import _FIELD_VALUE
from enhomie.homie.addons.persist import _FIELD_VALUE_LABEL
from enhomie.homie.addons.persist import _FIELD_VALUE_UNIT

from httpx import Response

if TYPE_CHECKING:
    from .plugin import HomiePlugin



class HomiePersistRecord(BaseModel, extra='ignore'):
    """
    Information relevant to Homie Automate persistent value.
    """

    unique: _FIELD_UNIQUE

    value: _FIELD_VALUE

    value_unit: _FIELD_VALUE_UNIT

    value_label: _FIELD_VALUE_LABEL

    about: _FIELD_ABOUT

    about_label: _FIELD_ABOUT_LABEL

    level: _FIELD_LEVEL

    tags: _FIELD_TAGS

    expire: _FIELD_EXPIRE

    update: _FIELD_UPDATE


    def __init__(
        self,
        /,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        super().__init__(**data)



class HomiePersist:
    """
    Collect the information using the upstream API endpoint.

    :param plugin: Plugin class instance for Chatting Robie.
    """

    __plugin: 'HomiePlugin'

    __client: HTTPClient

    __location: str


    def __init__(
        self,
        plugin: 'HomiePlugin',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__plugin = plugin

        params = plugin.params

        client = HTTPClient(
            timeout=params.timeout,
            verify=params.ssl_verify,
            capem=params.ssl_capem)

        self.__client = client

        self.__location = (
            f'{params.restful}'
            '/api/persists')


    @property
    def client(
        self,
    ) -> HTTPClient:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__client


    def request(
        self,
    ) -> Response:
        """
        Return the response from upstream API endpoint request.

        :param item: Determine which items will be enumerated.
        :returns: Response from upstream API endpoint request.
        """

        plugin = self.__plugin
        client = self.__client
        params = plugin.params
        location = self.__location

        request = client.request_block

        response = request(
            method='get',
            location=location,
            httpauth=(
                (params.username,
                 params.password)
                if params.username
                and params.password
                else None),
            timeout=params.timeout)

        (response
         .raise_for_status())


        return response


    @property
    def records(
        self,
    ) -> list[HomiePersistRecord]:
        """
        Return the response from upstream API endpoint request.

        :returns: Response from upstream API endpoint request.
        """

        returned: list[HomiePersistRecord] = []

        model = HomiePersistRecord


        response = self.request()


        items = (
            response.json()
            ['entries'])

        for item in items:

            object = model(**item)

            returned.append(object)


        return returned


    def record(
        self,
        unique: str,
    ) -> HomiePersistRecord:
        """
        Return the response from upstream API endpoint request.

        :param unique: Unique identifier from within the table.
        :returns: Response from upstream API endpoint request.
        """

        records = self.records


        for record in records:

            _unique = record.unique

            if _unique != unique:
                continue

            return record


        raise ValueError('unique')
